/****************************************************************)
(*								*)
(*								*)
(*	N       N        AAA       N       N      OOOOO		*)
(*	NN      N       A   A      NN      N     O     O	*)
(*	N N     N      A     A     N N     N    O       O	*)
(*	N  N    N     A       A    N  N    N    O       O	*)
(*	N   N   N     A       A    N   N   N    O       O	*)
(*	N    N  N     AAAAAAAAA    N    N  N    O       O	*)
(*	N     N N     A       A    N     N N    O       O	*)
(*	N      NN     A       A    N      NN     O     O	*)
(*	N       N     A       A    N       N      OOOOO		*)
(*								*)
(*								*)
(****************************************************************)
(*	Compiler fuer ein *sehr* eingeschraenktes Pascal.	*)
(*	Erzeugt Maschinencode fuer 8031 & co Prozessoren.	*)
(****************************************************************/

/*===================================================================*\
|*	$Id: code.c,v 1.1 2018/12/21 00:14:50 wolf Exp $
|*===================================================================*|
|*
|*	$Log: code.c,v $
|*	Revision 1.1  2018/12/21 00:14:50  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 22:27:22  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 16:03:56  wolf
|*	Initial revision
|*
|*	Revision 1.4  2011/08/09 17:26:17  wolf
|*	* Cleanup
|*
|*	Revision 1.3  2010/12/24 16:26:09  wolf
|*	*** empty log message ***
|*
|*	Revision 1.2  2008/12/12 20:22:21  wolf
|*	*** empty log message ***
|*
|*===================================================================*|
|*	1988-1996 als Turbopascal-Programm für 8080 und 8086
\*===================================================================*/


#include "nano.h"

#include "code.h"

#include "dasm.h"
#include "tree.h"
#include "hexfile.h"

#include <string.h>
#include <stdlib.h>

/*==============================================================*\
|*								*|
|*	Adressen						*|
|*								*|
\*==============================================================*/

extern uint16	Cadrs, Cadr, Tadr;

uint16
getCodeAddress (void) {
	return Cadr;
}

uint16
SetCodeAddress (uint16 addr) {
	uint16 oldcadr = Cadr;
	Cadr = addr;
	return oldcadr;
}


/*==============================================================*\
|*								*|
|*	Laenger der einzelnen Befehle				*|
|*								*|
\*==============================================================*/

uint8
CodeSize[256] = {
	1, 2, 3, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	3, 2, 3, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	3, 2, 1, 1, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	3, 2, 1, 1, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	2, 2, 2, 3, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	2, 2, 2, 3, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	2, 2, 2, 3, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	2, 2, 2, 1, 2, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
	2, 2, 2, 1, 1, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
	3, 2, 2, 1, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	2, 2, 2, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
	2, 2, 2, 1, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
	2, 2, 2, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	2, 2, 2, 1, 1, 3, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2,
	1, 2, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 2, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
};


/*==============================================================*\
|*								*|
|*	Elementare Codeerzeugung				*|
|*								*|
\*==============================================================*/

static const char sel[] = "code";

void
Code1 (uint8 opcode) {
	if (chkOption('l','c')) {
		lstfileselect (sel);
		fprintf (lstfile, "%04x' %02x      \t\t%s\n",
			Cadr, opcode,
			Dasm (Cadr, opcode, 0, 0));
	}
	Out (Cadr++, opcode);
	if (CodeSize[opcode]!=1) Error ("CodeSize1", 0);
}

void
Code2 (uint8 opcode, uint8 arg1) {
	if (chkOption('l','c')) {
		lstfileselect (sel);
		fprintf (lstfile, "%04x' %02x %02x   \t\t%s\n",
			Cadr, opcode, arg1,
			Dasm (Cadr, opcode, arg1, 0));
	}
	Out (Cadr++, opcode);
	Out (Cadr++, arg1);
	if (CodeSize[opcode]!=2) Error ("CodeSize2", 0);
}

void
Code3 (uint8 opcode, uint8 arg1, uint8 arg2) {
	if (chkOption('l','c')) {
		lstfileselect (sel);
		fprintf (lstfile, "%04x' %02x %02x %02x\t\t%s\n",
			Cadr, opcode, arg1, arg2, 
			Dasm (Cadr, opcode, arg1, arg2));
	}
	Out (Cadr++, opcode);
	Out (Cadr++, arg1);
	Out (Cadr++, arg2);
	if (CodeSize[opcode]!=3) Error ("CodeSize3", 0);
}

void
CodeByte (uint8 byte) {
	if (chkOption('l','c')) {
		lstfileselect (sel);
		fprintf (lstfile, "%04x' %02x      \t\tBYTE\t0x%02x\n",
			Cadr, byte, byte);
	}
	Out (Cadr++, byte);
}

void
CodeWord (uint16 word) {
	if (chkOption('l','c')) {
		lstfileselect (sel);
		fprintf (lstfile, "%04x' %02x %02x   \t\tWORD\t0x%04x\n",
			Cadr, lo(word), hi(word), word);
	}
	Out (Cadr++, lo(word));
	Out (Cadr++, hi(word));
}

uint16
CodeString (const char text[], unsigned len) {
	if (chkOption('l','c'))
		lstfileselect (sel);
	if (!len) len=strlen(text)+1;
	Tadr -= len;
	unsigned addr = Tadr;
	while (len-->0) {
		uint8 byte=*text++;
		if (chkOption('l','c'))
			fprintf (lstfile, "%04x' %02x      \t\tBYTE\t0x%02x\n",
				addr, byte, byte);
		Out (addr++, byte);
	}
	return Tadr;
}


/*==============================================================*\
|*								*|
|*	Spruenge und Calls zu bekannten Adressen		*|
|*								*|
\*==============================================================*/


/*--------------------------------------------------------------*\
|*	Ist das Sprungziel mit Adressierungsart erreichbar?	*|
\*--------------------------------------------------------------*/

static __inline BOOL
AOk (uint16 pc, uint16 target) {
	return !((pc^target) & 0xf800);
}


/*--------------------------------------------------------------*\
|*								*|
|*	Erzeuge Sprung-Befehl zu bekannter Adresse		*|
|*								*|
\*--------------------------------------------------------------*/

void
CodeJumpAddr (unsigned target, uint8 opcode, uint8 byte2) {

	/*------------------------------------------------------*\
	|*	Wenn nicht vorgegeben, wähle kürzesten Befehl	*|
	\*------------------------------------------------------*/

	if (!opcode) {
		opcode = 0x02;
		if (AOk(Cadr+2, target)) opcode=0x01;
		if (CodeIsRelativeOk(Cadr+2, target)) opcode=0x80;
	}

	/*------------------------------------------------------*\
	|*							*|
	\*------------------------------------------------------*/

	switch (opcode) {

	/*------------------------------------------------------*\
	|*	AJMP:	Sprung innerhalb einer 2k-Page		*|
	\*------------------------------------------------------*/

	case 0x01:

		if (!AOk(Cadr+2, target)) break;
		Code2 (((target>>3) & 0xe0) | 0x01, lo (target));
		return;

	/*------------------------------------------------------*\
	|*	LJMP:	Sprung auf beliebiges Ziel (3 byte)	*|
	\*------------------------------------------------------*/

	case 0x02:

		Code3 (0x02, hi(target), lo(target));
		return;

	/*------------------------------------------------------*\
	|*	JMPS:	Relative Sprünge ohne Zusatzbyte	*|
	\*------------------------------------------------------*/

	case 0x40: case 0x50:		/* Cy=1,  Cy=0	*/
	case 0x60: case 0x70:		/* Acc=0, Acc=1	*/
	case 0x80:			/* unbedingt	*/
	case 0xd8: case 0xd9:		/* DJNZ R0 / R1	*/
	case 0xda: case 0xdb:		/* DJNZ R2 / R3	*/
	case 0xdc: case 0xdd:		/* DJNZ R4 / R5	*/
	case 0xde: case 0xdf:		/* DJNZ R6 / R7	*/

		if (!CodeIsRelativeOk(Cadr+2, target)) break;
		Code2 (opcode, lo(target - (Cadr+2)));
		return;

	/*------------------------------------------------------*\
	|*	JMPS:	Relative Sprünge mit Bitnummer		*|
	\*------------------------------------------------------*/

	case 0x10: case 0x20: case 0x30:	/* JBC, JB, JNB	*/

	/*------------------------------------------------------*\
	|*	JMPS:	Relative Sprünge mit interner Adresse	*|
	\*------------------------------------------------------*/

	case 0xd5:			/* DJNZ        [mem]	*/
	case 0xb5:			/* CJNE A    , [mem]	*/

	/*------------------------------------------------------*\
	|*	JMPS:	Relative Sprünge mit Datenbyte		*|
	\*------------------------------------------------------*/

	case 0xb4:			/* CJNE A    , #c	*/
	case 0xb6: case 0xb7:		/* CJNE @R0/1, #c	*/
	case 0xb8: case 0xb9:		/* CJNE R0/R1, #c	*/
	case 0xba: case 0xbb:		/* CJNE R2/R3, #c	*/
	case 0xbc: case 0xbd:		/* CJNE R4/R5, #c	*/
	case 0xbe: case 0xbf:		/* CJNE R6/R7, #c	*/

		if (!CodeIsRelativeOk(Cadr+3, target)) break;
		Code3 (opcode, byte2, lo(target - (Cadr+3)));
		return;
	}

	/*------------------------------------------------------*\
	|*	Fehlerbehandlung				*|
	\*------------------------------------------------------*/

	fprintf (errfile, "PC=%04x TARGET=%04x\n", Cadr, target); /*FIXME*/
	Error (MsgInternal, "CodeJumpAddr");
}


/*--------------------------------------------------------------*\
|*								*|
|*	Erzeuge Call-Befehl zu bekannter Adresse		*|
|*								*|
\*--------------------------------------------------------------*/

void
CodeCallAddr (unsigned target, uint8 opcode) {

	/*------------------------------------------------------*\
	|*	Wenn nicht vorgegeben, wähle kürzesten Befehl	*|
	\*------------------------------------------------------*/

	if (!opcode) {
		opcode = 0x12;
		if (AOk(Cadr+2, target)) opcode=0x11;
	}

	/*------------------------------------------------------*\
	|*							*|
	\*------------------------------------------------------*/

	switch (opcode) {

	/*------------------------------------------------------*\
	|*	ACALL:	Call innerhalb einer 2k-Page		*|
	\*------------------------------------------------------*/

	case 0x11:

		if (!AOk(Cadr+2, target)) break;
		Code2 (((target>>3) & 0xe0) | 0x11, lo (target));
		return;

	/*------------------------------------------------------*\
	|*	LCALL:	Call auf beliebiges Ziel (3 byte)	*|
	\*------------------------------------------------------*/

	case 0x12:

		Code3 (0x12, hi(target), lo(target));
		return;

	/*------------------------------------------------------*\
	|*	Fehlerbehandlung				*|
	\*------------------------------------------------------*/

	}
	Error (MsgInternal, "CodeCallAddr");
}


/*==============================================================*\
|*								*|
|*	Label management fuer Spruenge vor und zurueck		*|
|*								*|
\*==============================================================*/


/*------------------------------------------------------*\
|*	Die erzeugten Label werden durchnummeriert	*|
\*------------------------------------------------------*/

unsigned LabelNumber=0;

/*------------------------------------------------------*\
|*	Der Disassmbler nutzt die Labelnummer		*|
\*------------------------------------------------------*/

Label * DasmLabel = 0;

/*------------------------------------------------------*\
|*	Verkettete Liste von nachzutragenden Sprüngen	*|
\*------------------------------------------------------*/

struct fixup {
	struct fixup*	next;
	uint16		faddr;
	uint8		opcode;
	uint8		byte2;
};


/*--------------------------------------------------------------*\
|*								*|
|*	Setze ein Label						*|
|*	- merke die Zieladresse					*|
|*	- erzeuge die zurückgestellten Sprünge zur Zieladresse	*|
|*								*|
\*--------------------------------------------------------------*/

void
CodeLabel (Label * label) {

	DasmLabel = label;

	/*------------------------------------------------------*\
	|*	Zieladresse des Labels ist "hier" :-)		*|
	\*------------------------------------------------------*/

	label->laddr = getCodeAddress();

	/*------------------------------------------------------*\
	|*	Ausgabe ins Listing				*|
	\*------------------------------------------------------*/

	if (chkOption('l','c')) {
		lstfileselect (sel);
		fprintf (lstfile, "%04x'         \tL%04u:\n",
			label->laddr, label->lnum);
	}

	/*------------------------------------------------------*\
	|*	Arbeite die Nachtragsliste ab			*|
	\*------------------------------------------------------*/

	struct fixup* fp;
	while ((fp=label->fixup)) {
		/*----------------------------------------------*\
		|*	Positioniere auf die Fixupadresse	*|
		\*----------------------------------------------*/
		SetCodeAddress (fp->faddr);
		/*----------------------------------------------*\
		|*	und erzeuge den Sprungbefehl		*|
		\*----------------------------------------------*/
		CodeJumpAddr (label->laddr, fp->opcode, fp->byte2);
		/*----------------------------------------------*\
		|*	entketten und struct entsorgen		*|
		\*----------------------------------------------*/
		label->fixup = fp->next;
		free (fp);
	}

	/*------------------------------------------------------*\
	|*	Codeerzeugung weiter ab "hier"			*|
	\*------------------------------------------------------*/

	SetCodeAddress (label->laddr);

	DasmLabel = NULL;
}

/*--------------------------------------------------------------*\
|*								*|
|*	Springe zu einem Label					*|
|*	- Wenn Zieladresse bekannt, erzeuge Sprungbefehl	*|
|*	- Sonst reserviere Platz für Sprungbefehl		*|
|*	  und erzeuge einen Eintrag in der Nachtragsliste
|*								*|
\*--------------------------------------------------------------*/

void
CodeJump (Label * label, uint8 opcode, uint8 byte2) {

	DasmLabel = label;

	if (label->laddr) {

		/*----------------------------------------------*\
		|*	Erzeuge den Sprungbefehl		*|
		\*----------------------------------------------*/

		CodeJumpAddr (label->laddr, opcode, byte2);

	} else {

		/*----------------------------------------------*\
		|*	Die CodeSize braucht einen fixen Opcode	*|
		\*----------------------------------------------*/

		if (!opcode) switch (getOption('J')) {
		case 1:	opcode = 0x80;
			break;
		case 2:	opcode = 0x01;
			break;
		default:
			opcode = 0x02;
		}

		/*----------------------------------------------*\
		|*	Erzeuge ein Nachtragslistenelement	*|
		\*----------------------------------------------*/

		struct fixup * fp = (struct fixup *) malloc(sizeof (struct fixup));

		/*----------------------------------------------*\
		|*	Fülle und verkette es			*|
		\*----------------------------------------------*/

		fp->next  = label->fixup;
		fp->faddr = Cadr;
		fp->opcode= opcode;
		fp->byte2 = byte2;
		label->fixup = fp;

		/*----------------------------------------------*\
		|*	Überspringe den Fixup-Bereich		*|
		\*----------------------------------------------*/

		Cadr += CodeSize[opcode];

		/*----------------------------------------------*\
		|*	Ausgabe in Listing			*|
		\*----------------------------------------------*/

		if (chkOption('l','c')) {
			lstfileselect (sel);
			fprintf (lstfile, "%04x' .. .. %s\t\t%s\n",
				fp->faddr,
				CodeSize[opcode]>=3 ? ".." : "  ",
				Dasm (Cadr, opcode, byte2, 0));
		}
	}

	DasmLabel = NULL;
}

/*--------------------------------------------------------------*\
|*								*|
|*	Erzeuge bedingten Sprung-Befehl 			*|
|*								*|
\*--------------------------------------------------------------*/

void
CodeCondJump (Label * label, BOOL on, uint8 ontrue, uint8 onfalse, uint8 byte2) {

	if (!label->laddr || CodeIsRelativeOk(Cadr+CodeSize[ontrue], label->laddr)) {
		CodeJump (label, on ? ontrue : onfalse, byte2);
		return;
	}

	Label skip;
	LabelInit (&skip);
	CodeJump (&skip, on ? onfalse:ontrue, byte2);
	CodeJump (label, 0, 0);
	CodeLabel (&skip);
}


/*===================================================================*\
|*	$Id: code.c,v 1.1 2018/12/21 00:14:50 wolf Exp $
\*===================================================================*/
