/****************************************************************)
(*								*)
(*								*)
(*	N       N        AAA       N       N      OOOOO		*)
(*	NN      N       A   A      NN      N     O     O	*)
(*	N N     N      A     A     N N     N    O       O	*)
(*	N  N    N     A       A    N  N    N    O       O	*)
(*	N   N   N     A       A    N   N   N    O       O	*)
(*	N    N  N     AAAAAAAAA    N    N  N    O       O	*)
(*	N     N N     A       A    N     N N    O       O	*)
(*	N      NN     A       A    N      NN     O     O	*)
(*	N       N     A       A    N       N      OOOOO		*)
(*								*)
(*								*)
(****************************************************************)
(*	Compiler fuer ein *sehr* eingeschraenktes Pascal.	*)
(*	Erzeugt Maschinencode fuer 8031 & co Prozessoren.	*)
(****************************************************************/

/*===================================================================*\
|*	$Id: code.h,v 1.1 2018/12/21 00:15:19 wolf Exp $
|*===================================================================*|
|*
|*	$Log: code.h,v $
|*	Revision 1.1  2018/12/21 00:15:19  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 22:27:37  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 16:03:56  wolf
|*	Initial revision
|*
|*	Revision 1.5  2011/08/09 17:26:58  wolf
|*	* Cleanup
|*
|*	Revision 1.4  2010/12/24 16:26:09  wolf
|*	*** empty log message ***
|*
|*	Revision 1.3  2008/12/12 20:22:21  wolf
|*	*** empty log message ***
|*
|*	Revision 1.2  2008/12/07 09:43:42  wolf
|*	*** empty log message ***
|*
|*===================================================================*|
|*	1988-1996 als Turbopascal-Programm für 8080 und 8086
\*===================================================================*/

#ifndef	__CODEGEN_H__
#define	__CODEGEN_H__

#include "config.h"

/*==============================================================*\
|*								*|
|*	Registerdefinitionen					*|
|*								*|
\*==============================================================*/

#define SP              0x81
#define DPL             0x82
#define DPH             0x83
#define PSW             0xd0
#define ACC             0xe0
#define B               0xf0
#define MD0             0xe9
#define MD1             0xea
#define MD2             0xeb
#define MD3             0xec
#define MD4             0xed
#define MD5             0xee
#define ARCON           0xef

#define DPC             0x100   /*  Code : @dptr  */
#define DPD             0x101   /*  Data : @dptr  */
#define R0D             0x102   /*  Data : @R0    */
#define R1D             0x103   /*  Data : @R1    */
		/*	0x104		constant  */
		/*	0x105		imemory   */
#define R0L             0x106   /*         @R0    */
#define R1L             0x107   /*         @R1    */
#define R0              0x108   /******************/
#define R1              0x109
#define R2              0x10a
#define R3              0x10b
#define R4              0x10c
#define R5              0x10d
#define R6              0x10e
#define R7              0x10f


/*==============================================================*\
|*								*|
|*	Elementare Coderezeugung				*|
|*								*|
\*==============================================================*/

uint16
getCodeAddress (void);

uint16
SetCodeAddress (uint16 addr);

void
Code1 (uint8 byte);

void
Code2 (uint8 byte, uint8 b2);

void
Code3 (uint8 byte, uint8 b2, uint8 b3);

void
CodeByte (uint8 byte);

void
CodeWord (uint16 word);

uint16
CodeString (const char text[], unsigned len);


/*==============================================================*\
|*								*|
|*	Spruenge vor und zurueck				*|
|*								*|
\*==============================================================*/

static inline uint8 lo (uint16 value) { return value & 0xff; }
static inline uint8 hi (uint16 value) { return (value>>8) & 0xff; }

typedef struct label {
	struct fixup*	fixup;
	uint16		laddr;
	uint16		lnum;
} Label;

static inline void
LabelInit (Label * label) {
	extern unsigned LabelNumber;
	label->fixup	= 0;
	label->laddr	= 0;
	label->lnum	= ++LabelNumber;
}

void
CodeLabel (Label * label);

void
CodeJumpAddr (unsigned target, uint8 opcode, uint8 byte2);

void
CodeCallAddr (unsigned target, uint8 opcode);

void
CodeJump (Label * label, uint8 opcode, uint8 byte2);

void
CodeCondJump (Label * label, BOOL on, uint8 optrue, uint8 opfalse, uint8 byte2);

/*--------------------------------------------------------------*\
|*	Ist das Sprungziel mit Adressierungsart erreichbar?	*|
\*--------------------------------------------------------------*/

static __inline BOOL
CodeIsRelativeOk (uint16 pc, uint16 target) {
	signed short dist = target - pc;
	return dist>=-128 && dist<=127;
}

extern uint8
CodeSize[256];

#endif

/*===================================================================*\
|*	$Id: code.h,v 1.1 2018/12/21 00:15:19 wolf Exp $
\*===================================================================*/
