/****************************************************************)
(*								*)
(*								*)
(*	N       N        AAA       N       N      OOOOO		*)
(*	NN      N       A   A      NN      N     O     O	*)
(*	N N     N      A     A     N N     N    O       O	*)
(*	N  N    N     A       A    N  N    N    O       O	*)
(*	N   N   N     A       A    N   N   N    O       O	*)
(*	N    N  N     AAAAAAAAA    N    N  N    O       O	*)
(*	N     N N     A       A    N     N N    O       O	*)
(*	N      NN     A       A    N      NN     O     O	*)
(*	N       N     A       A    N       N      OOOOO		*)
(*								*)
(*								*)
(****************************************************************)
(*	Compiler fuer ein *sehr* eingeschraenktes Pascal.	*)
(*	Erzeugt Maschinencode fuer 8031 & co Prozessoren.	*)
(****************************************************************/

/*===================================================================*\
|*	$Id: dasm.c,v 1.1 2018/12/21 00:14:50 wolf Exp $
|*===================================================================*|
|*
|*	$Log: dasm.c,v $
|*	Revision 1.1  2018/12/21 00:14:50  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 22:27:22  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 16:03:56  wolf
|*	Initial revision
|*
|*	Revision 1.4  2011/08/09 17:25:46  wolf
|*	* typo
|*
|*	Revision 1.3  2011/08/09 17:09:13  wolf
|*	Cleanup
|*
|*	Revision 1.3  2010/12/24 16:26:09  wolf
|*	*** empty log message ***
|*
|*	Revision 1.2  2008/12/12 20:22:21  wolf
|*	*** empty log message ***
|*
|*===================================================================*|
|*	1988-1996 als Turbopascal-Programm für 8080 und 8086
\*===================================================================*/

#include "dasm.h"

#include <string.h>
#include <stdio.h>


/*==============================================================*\
|*								*|
|*	Typen und Datenstrukturen				*|
|*								*|
\*==============================================================*/


/*------------------------------------------------------*\
|*	Parameter-Arten einer 8031-Instruktion		*|
\*------------------------------------------------------*/

typedef enum args {
	a,   ab,  ad,  ai, am, ams, ar,
	av,  avs, b,   bc, bs, c,   cb,  cq,
	d,   da,  dw,  i,  ia, im,  iv,
	ivs, j,   l,   m,  ma, mi,  mm,
	mr,  ms,  mv,  n,  r,  ra,  rm,
	rs,  rv,  rvs, s,  v,  xad, xap} OpArgs;


/*------------------------------------------------------*\
|*	Groesse der Instruktion zu jeder Parameter-Art	*|
\*------------------------------------------------------*/

static
const uint8 CodeSizeTable[] = {
	1,   1,   1,   1,  2,  3,   1,
	2,   3,   2,   2,  2,  1,   2,   2,
	1,   1,   3,   1,  1,  2,   2,
	3,   2,   3,   2,  2,  2,   3,
	2,   3,   3,   1,  1,  1,   2,
	2,   2,   3,   2,  2,  1,   1,
};


/*------------------------------------------------------*\
|*	Tabelle erstes Byte -> Opcode / Parameter-Art	*|
\*------------------------------------------------------*/

struct dasm {
	const char*	opcode;
	OpArgs		opargs;
};

static struct dasm DasmTable[256] = {
	{ "NOP"  , n	},
	{ "AJMP" , j	},
	{ "LJMP" , l	},
	{ "RR "  , a	},
	{ "INC"  , a	},
	{ "INC"  , m	},
	{ "INC"  , i	},
	{ "INC"  , i	},
	{ "INC"  , r	},
	{ "INC"  , r	},
	{ "INC"  , r	},
	{ "INC"  , r	},
	{ "INC"  , r	},
	{ "INC"  , r	},
	{ "INC"  , r	},
	{ "INC"  , r	},
	{ "JBC"  , bs	},
	{ "ACALL", j	},
	{ "LCALL", l	},
	{ "RRC"  , a	},
	{ "DEC"  , a	},
	{ "DEC"  , m	},
	{ "DEC"  , i	},
	{ "DEC"  , i	},
	{ "DEC"  , r	},
	{ "DEC"  , r	},
	{ "DEC"  , r	},
	{ "DEC"  , r	},
	{ "DEC"  , r	},
	{ "DEC"  , r	},
	{ "DEC"  , r	},
	{ "DEC"  , r	},
	{ "JB "  , bs	},
	{ "AJMP" , j	},
	{ "RET"  , n	},
	{ "RL "  , a	},
	{ "ADD"  , av	},
	{ "ADD"  , am	},
	{ "ADD"  , ai	},
	{ "ADD"  , ai	},
	{ "ADD"  , ar	},
	{ "ADD"  , ar	},
	{ "ADD"  , ar	},
	{ "ADD"  , ar	},
	{ "ADD"  , ar	},
	{ "ADD"  , ar	},
	{ "ADD"  , ar	},
	{ "ADD"  , ar	},
	{ "JNB"  , bs	},
	{ "ACALL", j	},
	{ "RETI" , n	},
	{ "RLC"  , a	},
	{ "ADDC"  , av	},
	{ "ADDC"  , am	},
	{ "ADDC"  , ai	},
	{ "ADDC"  , ai	},
	{ "ADDC"  , ar	},
	{ "ADDC"  , ar	},
	{ "ADDC"  , ar	},
	{ "ADDC"  , ar	},
	{ "ADDC"  , ar	},
	{ "ADDC"  , ar	},
	{ "ADDC"  , ar	},
	{ "ADDC"  , ar	},
	{ "JC "  , s	},
	{ "AJMP" , j	},
	{ "ORL"  , ma	},
	{ "ORL"  , mv	},
	{ "ORL"  , av	},
	{ "ORL"  , am	},
	{ "ORL"  , ai	},
	{ "ORL"  , ai	},
	{ "ORL"  , ar	},
	{ "ORL"  , ar	},
	{ "ORL"  , ar	},
	{ "ORL"  , ar	},
	{ "ORL"  , ar	},
	{ "ORL"  , ar	},
	{ "ORL"  , ar	},
	{ "ORL"  , ar	},
	{ "JNC"  , s	},
	{ "ACALL", j	},
	{ "ANL"  , ma	},
	{ "ANL"  , mv	},
	{ "ANL"  , av	},
	{ "ANL"  , am	},
	{ "ANL"  , ai	},
	{ "ANL"  , ai	},
	{ "ANL"  , ar	},
	{ "ANL"  , ar	},
	{ "ANL"  , ar	},
	{ "ANL"  , ar	},
	{ "ANL"  , ar	},
	{ "ANL"  , ar	},
	{ "ANL"  , ar	},
	{ "ANL"  , ar	},
	{ "JZ "  , s	},
	{ "AJMP" , j	},
	{ "XRL"  , ma	},
	{ "XRL"  , mv	},
	{ "XRL"  , av	},
	{ "XRL"  , am	},
	{ "XRL"  , ai	},
	{ "XRL"  , ai	},
	{ "XRL"  , ar	},
	{ "XRL"  , ar	},
	{ "XRL"  , ar	},
	{ "XRL"  , ar	},
	{ "XRL"  , ar	},
	{ "XRL"  , ar	},
	{ "XRL"  , ar	},
	{ "XRL"  , ar	},
	{ "JNZ"  , s	},
	{ "ACALL", j	},
	{ "ORL"  , cb	},
	{ "JMP"  , xad	},
	{ "MOV"  , av	},
	{ "MOV"  , mv	},
	{ "MOV"  , iv	},
	{ "MOV"  , iv	},
	{ "MOV"  , rv	},
	{ "MOV"  , rv	},
	{ "MOV"  , rv	},
	{ "MOV"  , rv	},
	{ "MOV"  , rv	},
	{ "MOV"  , rv	},
	{ "MOV"  , rv	},
	{ "MOV"  , rv	},
	{ "SJMP" , s	},
	{ "AJMP" , j	},
	{ "ANL"  , cb	},
	{ "MOVC" , xap	},
	{ "DIV"  , ab	},
	{ "MOV"  , mm	},
	{ "MOV"  , mi	},
	{ "MOV"  , mi	},
	{ "MOV"  , mr	},
	{ "MOV"  , mr	},
	{ "MOV"  , mr	},
	{ "MOV"  , mr	},
	{ "MOV"  , mr	},
	{ "MOV"  , mr	},
	{ "MOV"  , mr	},
	{ "MOV"  , mr	},
	{ "MOV"  , dw	},
	{ "ACALL", j	},
	{ "MOV"  , bc	},
	{ "MOVC" , ad	},
	{ "SUBB" , av	},
	{ "SUBB" , am	},
	{ "SUBB" , ai	},
	{ "SUBB" , ai	},
	{ "SUBB" , ar	},
	{ "SUBB" , ar	},
	{ "SUBB" , ar	},
	{ "SUBB" , ar	},
	{ "SUBB" , ar	},
	{ "SUBB" , ar	},
	{ "SUBB" , ar	},
	{ "SUBB" , ar	},
	{ "ORL"  , cq	},
	{ "AJMP" , j	},
	{ "MOV"  , cb	},
	{ "INC"  , d	},
	{ "MUL"  , ab	},
	{ "UNDEF", v	},
	{ "MOV"  , im	},
	{ "MOV"  , im	},
	{ "MOV"  , rm	},
	{ "MOV"  , rm	},
	{ "MOV"  , rm	},
	{ "MOV"  , rm	},
	{ "MOV"  , rm	},
	{ "MOV"  , rm	},
	{ "MOV"  , rm	},
	{ "MOV"  , rm	},
	{ "ANL"  , cq	},
	{ "ACALL", j	},
	{ "CPL"  , b	},
	{ "CPL"  , c	},
	{ "CJNE" , avs	},
	{ "CJNE" , ams	},
	{ "CJNE" , ivs	},
	{ "CJNE" , ivs	},
	{ "CJNE" , rvs	},
	{ "CJNE" , rvs	},
	{ "CJNE" , rvs	},
	{ "CJNE" , rvs	},
	{ "CJNE" , rvs	},
	{ "CJNE" , rvs	},
	{ "CJNE" , rvs	},
	{ "CJNE" , rvs	},
	{ "PUSH" , m	},
	{ "AJMP" , j	},
	{ "CLR"  , b	},
	{ "CLR"  , c	},
	{ "SWAP" , a	},
	{ "XCH"  , am	},
	{ "XCH"  , ai	},
	{ "XCH"  , ai	},
	{ "XCH"  , ar	},
	{ "XCH"  , ar	},
	{ "XCH"  , ar	},
	{ "XCH"  , ar	},
	{ "XCH"  , ar	},
	{ "XCH"  , ar	},
	{ "XCH"  , ar	},
	{ "XCH"  , ar	},
	{ "POP"  , m	},
	{ "ACALL", j	},
	{ "SETB" , b	},
	{ "SETB" , c	},
	{ "DA "  , a	},
	{ "DJNZ" , ms	},
	{ "XCD"  , ai	},
	{ "XCD"  , ai	},
	{ "DJNZ" , rs	},
	{ "DJNZ" , rs	},
	{ "DJNZ" , rs	},
	{ "DJNZ" , rs	},
	{ "DJNZ" , rs	},
	{ "DJNZ" , rs	},
	{ "DJNZ" , rs	},
	{ "DJNZ" , rs	},
	{ "MOVX" , ad	},
	{ "AJMP" , j	},
	{ "MOVX" , ai	},
	{ "MOVX" , ai	},
	{ "CLR"  , a	},
	{ "MOV"  , am	},
	{ "MOV"  , ai	},
	{ "MOV"  , ai	},
	{ "MOV"  , ar	},
	{ "MOV"  , ar	},
	{ "MOV"  , ar	},
	{ "MOV"  , ar	},
	{ "MOV"  , ar	},
	{ "MOV"  , ar	},
	{ "MOV"  , ar	},
	{ "MOV"  , ar	},
	{ "MOVX" , da	},
	{ "ACALL", j	},
	{ "MOVX" , ia	},
	{ "MOVX" , ia	},
	{ "CPL"  , a	},
	{ "MOV"  , ma	},
	{ "MOV"  , ia	},
	{ "MOV"  , ia	},
	{ "MOV"  , ra	},
	{ "MOV"  , ra	},
	{ "MOV"  , ra	},
	{ "MOV"  , ra	},
	{ "MOV"  , ra	},
	{ "MOV"  , ra	},
	{ "MOV"  , ra	},
	{ "MOV"  , ra	},
};


/*==============================================================*\
|*								*|
|*	Buffer fuer Disassembler Ausgabe			*|
|*								*|
\*==============================================================*/


static char dasmline[80];

char *p = dasmline;


/*------------------------------------------------------*\
|*	Haenge ein Zeichen an den Buffer an		*|
\*------------------------------------------------------*/

static inline void
addch (char c) { *p++ = c; *p=0; }


/*------------------------------------------------------*\
|*	Haenge einen String an den Buffer an		*|
\*------------------------------------------------------*/

static inline void
add (const char *src) { strcat(p, src); p+=strlen(src); }


/*==============================================================*\
|*								*|
|*	Haenge die unterschiedlichen Parameter an den Buffer an	*|
|*								*|
\*==============================================================*/


static void
arg_label (uint16 target) {
	extern Label*DasmLabel;
	if (DasmLabel && (DasmLabel->laddr==target || !DasmLabel->laddr))
		p+=snprintf (p, 10, "L%04u", DasmLabel->lnum);
	else
		p+=snprintf (p, 10, "0x%04x", target);
}

static inline void
arg_b (uint8 bit) {
	uint8 byte = bit & 0xf8;
	bit &= 0x07;
	if (byte<0x80) byte= byte/8+0x20;
	p+=snprintf (p, 12, "[0x%02x].%u", byte, bit);
}

static inline void
arg_i (uint8 code1) {
	add ("@R");
	addch ('0' + (code1 & 1));
}

static inline void
arg_j (uint8 code1, uint8 code2, uint16 addr) {
	arg_label((addr&0xf800) | ((code1<<3)&0x700) | code2);
}

static inline void
arg_l (uint8 hi, uint8 lo) {
	arg_label((hi<<8) | lo);
}

static inline void
arg_m (uint8 iaddr) {
	p+=snprintf (p, 12, "[0x%02x]", iaddr);
}

static inline void
arg_r (uint8 code1) {
	add ("R");
	addch ('0' + (code1 & 7));
}

static inline void
arg_s (uint8 offset, uint16 addr) {
	arg_label((uint16) ((short signed) addr + (signed char) offset));
}

static inline void
arg_v (uint8 data) {
	p+=snprintf (p, 10, "#0x%02x", data);
}

static inline void
arg_w (uint8 hi, uint8 lo) {
	p+=snprintf (p, 10, "#0x%02x%02x", hi, lo);
}


/*==============================================================*\
|*								*|
|*	Disassembler, einzige veroeffentlichte Routine		*|
|*								*|
\*==============================================================*/

const char *
Dasm (uint16 addr, uint8 byte1, uint8 byte2, uint8 byte3) {

	extern Label*DasmLabel;

	if (DasmLabel && DasmLabel->laddr && DasmLabel->fixup)
		return "FIXUP";

	p = dasmline;
	p[0]=0;

	add (DasmTable[byte1].opcode);
	add ("\t");

	switch (DasmTable[byte1].opargs) {

	case a:
		add ("A");
		break;
	case ab:
		add ("AB");
		break;
	case ad:
		add ("A,@DPTR");
		break;
	case ai:
		add ("A,");
		arg_i(byte1);
		break;
	case am:
		add ("A,");
		arg_m(byte2);
		break;
	case ams:
		add ("A,");
		arg_m(byte2);
		add (",");
		arg_s(byte3, addr+3);
		break;
	case ar:
		add ("A,");
		arg_r(byte1);
		break;
	case av:
		add ("A,");
		arg_v(byte2);
		break;
	case avs:
		add ("A,");
		arg_v(byte2);
		add (",");
		arg_s(byte3, addr+3);
		break;
	case b:
		arg_b(byte2);
		break;
	case bc:
		arg_b(byte2);
		add (",C");
		break;
	case bs:
		arg_b (byte2);
		add   (",");
		arg_s (byte3, addr+3);
		break;
	case c:
		add ("C");
		break;
	case cb:
		add ("C,");
		arg_b(byte2);
		break;
	case cq:
		add ("C,¬");
		arg_b(byte2);
		break;
	case d:
		add ("DPTR");
		break;
	case da:
		add ("@DPTR,A");
		break;
	case dw:
		add ("DPTR,");
		arg_w(byte2,byte3);
		break;
	case i:
		arg_i(byte1);
		break;
	case ia:
		arg_i(byte1);
		add (",A");
		break;
	case im:
		arg_i(byte1);
		add (",");
		arg_m(byte2);
		break;
	case iv:
		arg_i(byte1);
		add (",");
		arg_v (byte2);
		break;
	case ivs:
		arg_i (byte1);
		add   (",");
		arg_v (byte2);
		add   (",");
		arg_s (byte3, addr+3);
		break;
	case j:
		arg_j (byte1, byte2, addr+2);
		break;
	case l:
		arg_l (byte2, byte3);
		break;
	case m:
		arg_m (byte2);
		break;
	case ma:
		arg_m (byte2);
		add   (",A");
		break;
	case mi:
		arg_m (byte2);
		add   (",");
		arg_i (byte1);
		break;
	case mm:
		arg_m (byte3);
		add   (",");
		arg_m (byte2);
		break;
	case mr:
		arg_m (byte2);
		add   (",");
		arg_r (byte1);
		break;
	case ms:
		arg_m (byte2);
		add   (",");
		arg_s (byte3, addr+3);
		break;
	case mv:
		arg_m (byte2);
		add   (",");
		arg_v (byte3);
		break;
	case n:
		break;
	case r:
		arg_r (byte1);
		break;
	case ra:
		arg_r (byte1);
		add   (",A");
		break;
	case rm:
		arg_r (byte1);
		add   (",");
		arg_m (byte2);
		break;
	case rs:
		arg_r (byte1);
		add   (",");
		arg_s (byte2, addr+2);
		break;
	case rv:
		arg_r (byte1);
		add   (",");
		arg_v (byte2);
		break;
	case rvs:
		arg_r (byte1);
		add   (",");
		arg_v (byte2);
		add   (",");
		arg_s (byte3, addr+3);
		break;
	case s:
		arg_s (byte2, addr+2);
		break;
	case v:
		arg_v (byte1);
		break;
	case xad:
		add   ("@A+DPTR");
		break;
	case xap:
		add   ("@A+PC");
		break;
	default:
		add   ("???");
	}

	*p = 0;
	return dasmline;
}


/*===================================================================*\
|*	$Id: dasm.c,v 1.1 2018/12/21 00:14:50 wolf Exp $
\*===================================================================*/
