/****************************************************************)
(*								*)
(*								*)
(*	N       N        AAA       N       N      OOOOO		*)
(*	NN      N       A   A      NN      N     O     O	*)
(*	N N     N      A     A     N N     N    O       O	*)
(*	N  N    N     A       A    N  N    N    O       O	*)
(*	N   N   N     A       A    N   N   N    O       O	*)
(*	N    N  N     AAAAAAAAA    N    N  N    O       O	*)
(*	N     N N     A       A    N     N N    O       O	*)
(*	N      NN     A       A    N      NN     O     O	*)
(*	N       N     A       A    N       N      OOOOO		*)
(*								*)
(*								*)
(****************************************************************)
(*	Compiler fuer ein *sehr* eingeschraenktes Pascal.	*)
(*	Erzeugt Maschinencode fuer 8031 & co Prozessoren.	*)
(****************************************************************/

/*===================================================================*\
|*	$Id: declarations.c,v 1.1 2018/12/21 00:14:50 wolf Exp $
|*===================================================================*|
|*
|*	$Log: declarations.c,v $
|*	Revision 1.1  2018/12/21 00:14:50  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 22:27:22  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 16:03:56  wolf
|*	Initial revision
|*
|*	Revision 1.4  2011/08/09 17:26:17  wolf
|*	* Cleanup
|*
|*	Revision 1.3  2010/12/24 16:26:09  wolf
|*	*** empty log message ***
|*
|*	Revision 1.2  2008/12/12 20:22:21  wolf
|*	*** empty log message ***
|*
|*	Revision 1.1  2008/12/07 09:43:42  wolf
|*	Initial revision
|*
|*===================================================================*|
|*	1988-1996 als Turbopascal-Programm für 8080 und 8086
\*===================================================================*/

#include "nano.h"

#include "symboltable.h"

#include "scanner.h"
#include "expressions.h"
#include "statements.h"

#include "memory.h"
#include "code.h"

#include <stdio.h>


/********************************************************\
|*							*|
|*	Hilfsroutinen fuer die Deklarationen		*|
|*							*|
\********************************************************/


/*---------------------------------------------*\
|						|
|	Absolute Adresse [constexpression]	|
|						|
\*---------------------------------------------*/

static void
DoAbsoluteAddress (Symbol sym) {

	if (getToken() != seka) return;
	Consume(seka);

	if (getToken() == sident) {
		Consume (sident);
		Symbol oldsym= FindSymbol(getTokenName());
		sym->address = oldsym->address;
		sym->storage = oldsym->storage;
	} else {
		sym->address = ConstantValueOfType(word);
	}
	sym->absolute = TRUE;
	Consume(sekz);
}


/*---------------------------------------------*\
|						|
|	Typangabe	(byte, word, string)	|
|						|
\*---------------------------------------------*/

static void
DoType (Symbol sym) {

	switch (getToken()) {

	case sbyte:
		sym->type = byte;
		break;

	case sword:
		sym->type = word;
		break;

	case sstring:
		sym->type = string;
		break;

	default:
		Error (MsgExpectedTypename, 0);
		break;
	}
	Cns();
}


/*==============================================================*\
|*								*|
|*								*|
|*		Elementare Datendeklarationen			*|
|*								*|
|*								*|
\*==============================================================*/


/*---------------------------------------------*\
|						|
|	CONST id = <constexpression>		|
|						|
\*---------------------------------------------*/

static void
DoConst (void) {
	Consume(sconst);
	Consume(sident);
	Symbol sym = CreateSymbol (constant, getTokenName());
	Consume(seq);
	sym->address = ConstantValue (&sym->type);
	sym->storage = sym->type==string ? code : non;
	Consume(ssemi);
}


/*---------------------------------------------------------------------*\
|									|
|	VAR id [absadr]: [ARRAY <const> OF] <typ> [INTERN|EXTERN]	|
|	VAR id [absadr]: (enum0, enum1, enum2, ... enumn)		|
|	VAR id [absadr]: [b7 b6 b5 b4 b3 b2 b2 b0]			|
|									|
\*---------------------------------------------------------------------*/

static void
DoVariable (void) {

	Consume(svar);
	Consume(sident);

	Symbol sym = CreateSymbol (variable, getTokenName());
	sym->storage = internal;
	DoAbsoluteAddress(sym);
	Consume(scolon);

	/*-------------------------------------------------------------*\
	|	VAR id [absadr]: (enum0, enum1, enum2, ... enumn)	|
	|---------------------------------------------------------------|
	|	Definiert neben einer internen Bytevariable		|
	|	zusaetzlich die Konstanten enum0=0, enum1=1, ...	|
	\*-------------------------------------------------------------*/

	if (getToken() == srka) {
		sym->type = byte;
		AllocateMemory(sym);
		unsigned value = 0;
		do {
			Cns();
			Consume(sident);
			Symbol enumsym = CreateSymbol (constant, getTokenName());
			enumsym->address = value++;
			enumsym->type = word;
		} while (getToken() == scomma);
		Consume(srkz);
		Consume(ssemi);
		return;
	}

	/*-------------------------------------------------------------*\
	|	VAR id [absadr]: [b7 b6 b5 b4 b3 b2 b1 b0]		|
	|---------------------------------------------------------------|
	|	Definiert neben einer internen Bytevariable		|
	|	zusaetzlich die Bitvariablen b0..b7.			|
	|	Nicht benötigte können als "*" angegeben werden		|
	|	Ein "/" vor einem Bitnamen invertiert dieses Bit	|
	\*-------------------------------------------------------------*/

	if (getToken() == seka) {
		sym->type = byte;
		AllocateMemory(sym);
		if (sym->absolute && sym->address>=0x80) sym->storage=reg;
		uint8 basead = sym->address;
		Cns();
		int bitnum;
		for (bitnum=7; bitnum>=0; bitnum--) {
			if (getToken() == smul) {
				Cns();
				continue;
			}
			uint16 invert = 0;
			if (getToken() == sexclam) {
				Cns();
				invert = 0x8000;
			}
			Consume(sident);
			Symbol bitsym = CreateSymbol (bitvariable, getTokenName());
			bitsym->address = basead + (bitnum << 8) + invert;
			bitsym->storage = sym->storage;
			bitsym->type    = bit;
		}
		Consume(sekz);
		Consume(ssemi);
		return;
	}

	/*-------------------------------------------------------------*\
	|	VAR id [absadr]: ARRAY elements OF <type>		|
	|---------------------------------------------------------------|
	|	Angegeben wird die nur die Groesse des Arrays		|
	|	Die Indizierung beginnt immer mit 0.			|
	|	Array werden per default im oberen Teil des internen	|
	|	Speichers angelegt.					|
	\*-------------------------------------------------------------*/

	if (getToken() == sarray) {
		if (sym->storage==internal) sym->storage = upb;
		Consume(sarray);
		sym->size = ConstantValueOfType(word);
		Consume(sof);
	}

	DoType(sym);

	/*-------------------------------------------------------------*\
	|	Default-Speichersegment ist "internal"			|
	|	und zwar oberer Teil (=udb) für Arrays,			|
	|	und unterer Teil (=default) für Skalarvariablen		|
	|---------------------------------------------------------------|
	|	Bei Variablen mit absoluter Adresse wird das Segment	|
	|	von der Referenzvariablen übernommen			|
	|	Mit angehaengten Schluesselwoertern kann das Segment	|
	|	explizit gewäehlt werden: INTERN, EXTERN		|
	\*-------------------------------------------------------------*/

	if (getToken() == sintern) {
		sym->storage = internal;
		Cns();
	} else if (getToken() == sextern) {
		sym->storage = external;
		Cns();
	}

	AllocateMemory(sym);
	Consume(ssemi);
}



/*---------------------------------------------------------------------*\
|									|
|	TABLE id [absadr]: ARRAY <const> OF <typ> = [ v1, v2 ... ]	|
|									|
|-----------------------------------------------------------------------|
|	Table erzeugt ein initialisiertes und nicht änderbares Array	|
|	Die Table wird im Codespeicher abgelegt.			|
\*---------------------------------------------------------------------*/

static void
DoTable(void) {

	Consume(stable);
	Consume(sident);

	Symbol sym = CreateSymbol (variable, getTokenName());

	DoAbsoluteAddress(sym);

	/*-------------------------------------------------------------*\
	|	Bestimme die Groesse der Tabelle			|
	\*-------------------------------------------------------------*/

	Consume(scolon);
	Consume(sarray);
	sym -> size = ConstantValueOfType(word);
	Consume(sof);

	/*-------------------------------------------------------------*\
	|	Bestimme den Typ der Tabelle				|
	\*-------------------------------------------------------------*/

	DoType(sym);

	Consume(seq);
	Consume(seka);

	/*-------------------------------------------------------------*\
	|	Gespeichert wird im Codesegment (@Cadr)			|
	|	Bei absoluter Adresse muß Cadr vorher gesetzt werden	|
	\*-------------------------------------------------------------*/

	sym -> storage = code;
	uint16 savecadr=0;
	if (sym->absolute) {
		savecadr= SetCodeAddress (sym->address);
	} else {
		sym->address = getCodeAddress();
	}

	/*-------------------------------------------------------------*\
	|	Nun hole die Konstanten für die Tabelle...		|
	\*-------------------------------------------------------------*/

	unsigned count = sym -> size;
	while (count>0) {

		uint16 value = ConstantValue (&sym->type);

		/*-------------------------------------------------------------*\
		|	<const1> : <times>					|
		|	repliziert die erste Konstante				|
		\*-------------------------------------------------------------*/

		unsigned repeat = 1;
		if (getToken() == scolon) {
			Consume (scolon);
			repeat = ConstantValueOfType(word);
		}

		/*-------------------------------------------------------------*\
		|	Speichere die Werte					|
		\*-------------------------------------------------------------*/

		while (repeat>0 && count>0) {
			repeat--;
			count --;
			if (sym->type == byte) {
				CodeByte (lo (value));
			} else {
				CodeWord (value);
			}
		}

		/*-------------------------------------------------------------*\
		|	Erwarte und verbrauchen ein Komma			|
		\*-------------------------------------------------------------*/

		if (count>0) Consume (scomma);

	}

	/*-------------------------------------------------------------*\
	|	Bei absoluter Adresse muß Cadr restauriert werden	|
	\*-------------------------------------------------------------*/

	if (sym->absolute) SetCodeAddress (savecadr);

	Consume(sekz);
	Consume(ssemi);
}


/*##############################################################*\
|#								#|
|#								#|
|#		Prozeduren, Interrupts usw.			#|
|#								#|
|#								#|
\*##############################################################*/


static void
DoBlock (Scope scope, Symbol sym, int options);


/*---------------------------------------------------------------------*\
|									|
|	PROCEDURE id ([VAR] id [<absadr>]: [ARRAY OF] typ, ...)		|
|									|
|-----------------------------------------------------------------------|
|	Die Parameter der Prozeduren werden nicht auf dem Stack,	|
|	sondern in "normalen Variablen" uebergeben.			|
|	Daher keine Rekursion.						|
\*---------------------------------------------------------------------*/

static void
DoProcedure(void) {

	Consume(sprocedure);
	Consume(sident);

	Symbol sym = CreateSymbol (noclass, getTokenName());
	sym->type    = proc;

	unsigned vector = 0;

	if (getToken() == seka) {
		Consume(seka);
		vector = ConstantValueOfType(word);
		Consume(sekz);
	}

	if (getToken() == srka) {

		/*-------------------------------------*\
		|	Optionale Parameterliste	|
		\*-------------------------------------*/

		Consume (srka);

		for (;;) {

			/*-------------------------------------*\
			|	size zaehlt die Argument 	|
			\*-------------------------------------*/

			sym->size++;

			/*-------------------------------------*\
			|	Uebergabe "by reference" ? 	|
			\*-------------------------------------*/

			BOOL byref=FALSE;
			if (getToken()==svar) {
				Consume (svar);
				byref=TRUE;
			}

			/*-------------------------------------*\
			|	Name des Parameters		|
			\*-------------------------------------*/

			Consume(sident);
			Symbol parasym = CreateSymbol (variable, getTokenName());

			/*-------------------------------------*\
			|	Absolute Adresse fuer Parameter	|
			\*-------------------------------------*/

			parasym->indirect = byref;
			DoAbsoluteAddress(parasym);
			Consume(scolon);

			/*-------------------------------------*\
			|	FIXME voellig ungetestet	|
			\*-------------------------------------*/

			if (getToken() == stable) {
				Cns();
				parasym->storage = code;
				parasym->size = ConstantValueOfType(word);
				Consume(sof);
			} else if (getToken() == sarray) {
				Cns();
				parasym->storage = external;
				parasym->size = ConstantValueOfType(word);
				Consume(sof);
			}

			/*-------------------------------------*\
			|	Typ des Parameters		|
			\*-------------------------------------*/

			DoType(parasym);

			/*-------------------------------------*\
			|	Alloziere (statisch) Speicher	|
			\*-------------------------------------*/

			AllocateMemory(parasym);

			/*-------------------------------------*\
			|	Mehr Parameter?			|
			\*-------------------------------------*/

			if (getToken() == srkz) break;

			Consume (ssemi);
		}

		Consume(srkz);
	}

	Consume(ssemi);

	/*-------------------------------------*\
	|	Deklarationen und Code		|
	\*-------------------------------------*/

	DoBlock (procedure, sym, 0);
	Consume(ssemi);

	/*------------------------------------------------------*\
	|*	Sprung auf aktuell Adresse in den Vektor	*|
	\*------------------------------------------------------*/

	if (vector) {
		uint16 oldcadr = SetCodeAddress(vector);
		CodeJumpAddr (sym->address, 0, 0);
		SetCodeAddress (oldcadr);
		sym->used=TRUE;
	}

	/*------------------------------------------------------*\
	|*	Datenobjekte loeschen und anzeigen		*|
	\*------------------------------------------------------*/

	ReleaseScope (sym, sym->size);
}


/*---------------------------------------------------------------------*\
|									|
|	INTERRUPT   <num>  [ [bank] ]; <block>				|
|	INTERRUPT @ <addr> [ [bank] ]; <block>				|
|									|
|-----------------------------------------------------------------------|
|	Auf die angegebene oder berechne Vektoradresse wird ein		|
|	Sprung zum Interrupt-Handler abgelegt.				|
|	Der Interrupthandler sichert die Register PSW,A,B,DPW		|
|	Wenn eine Registerbank (1..3) angegeben ist, so wird danach	|
|	das PSW so gesetzt, daß es diese Registerbank auswaehlt.	|
|	Registerbank 0 unterdrückt das Retten der Register		|
\*---------------------------------------------------------------------*/


static void
DoInterrupt(void) {

	Consume(sinterrupt);

	/*------------------------------------------------------*\
	|*	Bestimme die Adresse für den Sprungvector	*|
	\*------------------------------------------------------*/

	uint16 vector=0;

	if (getToken()==sat) {
		Consume(sat);
		vector = ConstantValueOfType(word);
	} else {
		uint16 number = ConstantValueOfType(word) - 1;
		vector = number*8 + 3;
		static unsigned InterruptUsed;
		if (number>=20 || (number&7) >= 6 || vector+2 >= Cadrs || (InterruptUsed & (1 << number)))
			Error (MsgInvalid, "interrupt vector");
		InterruptUsed |= 1<<number;
	}

	/*------------------------------------------------------*\
	|*	Bestimme die Registersicherung			*|
	|*	bank != 0:	PSW ACC DPl DPW B retten	*|
	|*	bank >= 1:	zus. Registerbank schalten	*|
	\*------------------------------------------------------*/

	Scope scope = interrupt;
	uint8	psw = 0;

	if (getToken() == seka) {
		Consume(seka);
		uint16 bank = ConstantValueOfType(word);
		if (bank>3) Error (MsgInvalid, "register bank");
		Consume(sekz);

		if (bank>0) {
			psw = bank << 8;
		} else {
			scope = fastinterrupt;
		}
	}

	/*------------------------------------------------------*\
	|*	Definiere ein Symbol
	\*------------------------------------------------------*/

	char id[24];
	snprintf (id, 24, "<interrupt:%04x>", vector);
	Symbol sym = CreateSymbol (noclass, id);

	/*------------------------------------------------------*\
	|*	Und der Interrupt-Body				*|
	\*------------------------------------------------------*/

	Consume(ssemi);
	DoBlock(scope, sym, psw);
	Consume(ssemi);

	/*------------------------------------------------------*\
	|*	Sprung auf aktuell Adresse in den Vektor	*|
	\*------------------------------------------------------*/

	uint16 oldcadr = SetCodeAddress((Cadrs & 0x8000) | vector);
	CodeJumpAddr (sym->address, 0, 0);
	SetCodeAddress (oldcadr);
	sym->used=TRUE;

	/*------------------------------------------------------*\
	|*	Datenobjekte loeschen und anzeigen		*|
	\*------------------------------------------------------*/

	ReleaseScope (sym, 0);
}


/*##############################################################*\
|#								#|
|#								#|
|#		Deklarationen und Bloecke			#|
|#								#|
|#								#|
\*##############################################################*/


/*---------------------------------------------------------------------*\
|									|
|	Declarations:							|
|		CONST  id = val						|
|		VAR    id : type					|
|		TABLE  id : ARRAY n IF type				|
|		PROCEDURE id (para); <decl> BEGIN <stmtseq> END		|
|		INTERRUPT [@]n [bank];      BEGIN <stmtseq> END		|
|		INLINE    addr              BEGIN <stmtseq> END		|
|									|
\*---------------------------------------------------------------------*/

static void
DoDeclarations (void) {

	for (;;) {

		switch (getToken()) {

		case sconst:
			DoConst();
			continue;

		case svar:
			DoVariable();
			continue;

		case stable:
			DoTable();
			continue;

		case sprocedure:
			DoProcedure();
			continue;

		case sinterrupt:
			DoInterrupt();
			continue;

		default:
			break;
		}
		return;
	}
}


/*---------------------------------------------------------------------*\
|									|
|	Block:								|
|		<decl1>; <decl2>; .. <decln>				|
|		[BEGIN <stmtseq>]					|
|		END							|
|									|
\*---------------------------------------------------------------------*/


static void
DoBlock (Scope scope, Symbol sym, int options) {

	/*-------------------------------------*\
	|	Declarations			|
	\*-------------------------------------*/

	DoDeclarations();

	ThisScope = scope;
	sym->address = getCodeAddress();
	sym->storage = code;

	/*-------------------------------------*\
	|	Body				|
	\*-------------------------------------*/

	if (getToken() == sbegin) {
		Cns ();
    		DoEnter (options);
    		DoStatementSequence (TRUE, 0, 0);
		DoLeave();
	} else {
		DoLeaveImmediately();
	}
	Consume(send);
}


/*---------------------------------------------------------------------*\
|									|
|	Program:							|
|		PROGRAM id [ ( memoryranges ) ]; 			|
|		<decl1>; <decl2>; .. <decln>				|
|		[BEGIN <stmtseq>]					|
|		END							|
|									|
|	memoryranges:							|
|		(							|
|		[codestart	codeend]				|
|		[externalstart	externalend]				|
|		[internamstart	internalend]				|
|									|
|									|
\*---------------------------------------------------------------------*/


void
DoProgram (void) {

	Consume(sprogram);
	Consume(sident);

	Symbol sym = CreateSymbol (noclass, getTokenName());
	sym->type    = notype;

	/*------------------------------------------------------*\
	|*	Speichersegmente definieren			*|
	\*------------------------------------------------------*/

	if (getToken() == srka) {
		Consume(srka);

		Consume(seka);
		Cadrs = ConstantValueOfType(word);
		Consume(supto);
		Cadre = ConstantValueOfType(word);
		Consume(sekz);

		Consume(seka);
		Dadrs = ConstantValueOfType(word);
		Consume(supto);
		Dadre = ConstantValueOfType(word);
		Consume(sekz);

		Consume(seka);
		Ladrs = ConstantValueOfType(word);
		Consume(supto);
		Uadre = ConstantValueOfType(word);
		Consume(sekz);

		Consume(srkz);

		MemoryReset();
	}

	/*------------------------------------------------------*\
	|*	Deklarationen und Befehle			*|
	\*------------------------------------------------------*/

	Consume (ssemi);
	DoBlock (program, sym, 0);
	Consume (sdot);

	/*------------------------------------------------------*\
	|*	Sprung auf aktuell Adresse in den Vektor	*|
	\*------------------------------------------------------*/

	uint16 oldcadr = SetCodeAddress (Cadrs & 0x8000);
	CodeJumpAddr (sym->address, 0, 0);
	SetCodeAddress (oldcadr);
	sym->used=TRUE;

	/*------------------------------------------------------*\
	|*	Datenobjekte loeschen und anzeigen		*|
	\*------------------------------------------------------*/

	ReleaseScope (sym, 0);
}


/*===================================================================*\
|*	$Id: declarations.c,v 1.1 2018/12/21 00:14:50 wolf Exp $
\*===================================================================*/
