/****************************************************************)
(*								*)
(*								*)
(*	N       N        AAA       N       N      OOOOO		*)
(*	NN      N       A   A      NN      N     O     O	*)
(*	N N     N      A     A     N N     N    O       O	*)
(*	N  N    N     A       A    N  N    N    O       O	*)
(*	N   N   N     A       A    N   N   N    O       O	*)
(*	N    N  N     AAAAAAAAA    N    N  N    O       O	*)
(*	N     N N     A       A    N     N N    O       O	*)
(*	N      NN     A       A    N      NN     O     O	*)
(*	N       N     A       A    N       N      OOOOO		*)
(*								*)
(*								*)
(****************************************************************)
(*	Compiler fuer ein *sehr* eingeschraenktes Pascal.	*)
(*	Erzeugt Maschinencode fuer 8031 & co Prozessoren.	*)
(****************************************************************/

/*===================================================================*\
|*	$Id: expressions.c,v 1.2 2019/08/03 00:05:00 wolf Exp $
|*===================================================================*|
|*
|*	$Log: expressions.c,v $
|*	Revision 1.2  2019/08/03 00:05:00  wolf
|*	!!!!! Bugfix: 0-Byte am String-Ende fehlte.
|*
|*	Revision 1.1  2018/12/21 00:14:50  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 22:27:22  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 16:03:56  wolf
|*	Initial revision
|*
|*	Revision 1.3  2011/08/09 17:26:17  wolf
|*	* Cleanup
|*
|*	Revision 1.2  2008/12/12 20:22:21  wolf
|*	*** empty log message ***
|*
|*===================================================================*|
|*	1988-1996 als Turbopascal-Programm für 8080 und 8086
\*===================================================================*/


#include "nano.h"

#include "expressions.h"

#include "code.h"
#include "scanner.h"
#include "symboltable.h"
#include "tree.h"

#include <stdio.h>
#include <string.h>


/*==============================================================*\
|*								*|
|*	Handling von Typen und Speicherklassen			*|
|*								*|
\*==============================================================*/


/*----------------------------------------------*\
|*	Garantiere einen bestimmten Typ		*|
\*----------------------------------------------*/

void
VerifyType (Type found, Type expected) {
	if (found==expected) return;
	Error (MsgExpectedType, expected==string ? "string" : "numeric");
}


/*----------------------------------------------*\
|*	Bestimme Groesse in Bytes eines Typs	*|
\*----------------------------------------------*/

uint16
typesize (Type t) {
	return t==byte ? 1 : 2;
}


/*------------------------------------------------------*\
|*	Bestimme die Adressgroesse für Speicherklasse	*|
\*------------------------------------------------------*/

uint16
address_size (Storage s) {
	return s==internal ? 1 : 2;
}


/*==============================================================*\
|*								*|
|*	Parse Variable incl. Indizierung und .LO/.HI-Suffix	*|
|*								*|
\*==============================================================*/

uint16
GetAddress (Symbol sym) {

	/*----------------------------------------------*\
	|*	Bestehe auf einer Variablen		*|
	\*----------------------------------------------*/

	if (sym->sclass != variable)
		Fatal (MsgExpectedVariable, 0);

	/*----------------------------------------------*\
	|*	Bearbeite Indizierung			*|
	\*----------------------------------------------*/

	BOOL index = sym->size>0 && (getToken()==seka);

	unsigned asize = address_size(sym->storage);
	unsigned elementsize = typesize(sym->type);

	if (index) {
		Consume(seka);
		GetValue(word);
		Consume(sekz);

		switch (elementsize) {
		case 1:
			break;
		case 2:
			CreateNode (shl1u, 0, asize, FALSE);
			break;
		default:
			Error (MsgInternal, "GetAddress:elementsize");
		}
	}

	/*----------------------------------------------*\
	|*	Knoten mit Adresse			*|
	\*----------------------------------------------*/

	CreateConstantNode (sym->address);

	/*----------------------------------------------*\
	|*	Indirekte Adressierung (VAR-Parameter)	*|
	\*----------------------------------------------*/

	if (sym->indirect) CreateLoadNode (internal, asize);

	/*----------------------------------------------*\
	|*	Bearbeite ".lo" und ".hi" Suffixe	*|
	\*----------------------------------------------*/

	if (sym->type==word && getToken()== sdot) {

		Cns();
		if (getToken()==shi) {
			Cns();
			CreateConstantNode (1);
			CreateNode (addu, asize, 0, FALSE);
		} else {
			Consume(slo);
		}
		elementsize = 1;
	}

	/*----------------------------------------------*\
	|*	Addiere den Index dazu, wenn nötig	*|
	\*----------------------------------------------*/

	if (index) CreateNode (addu, asize, 0, FALSE);

	/*----------------------------------------------*\
	|*	Im Baum liegt die aktuelle Adresse,	*|
	|*	zurueckgeliefert wird die Groesse.	*|
	\*----------------------------------------------*/

	return elementsize;
}


/*==============================================================*\
|*								*|
|*								*|
|*		Elementarausdruecke				*|
|*								*|
|*								*|
\*==============================================================*/

static Type
DoIdent (void) {

	Type t;

	Consume (sident);
	Symbol sym = FindSymbol (getTokenName());

	if (sym->sclass == constant) {

		CreateConstantNode (sym->address);
		t = sym->type;

	} else if (sym->sclass == variable) {

		unsigned size = GetAddress (sym);

		CreateLoadNode (sym->storage, size);

		t = sym->type;
		if (t==byte) t=word;

	} else if (sym->sclass == bitvariable) {

		CreateConstantNode (sym->address & 0xff);	/* byte@ */
		CreateLoadNode (sym->storage, 1);
		CreateConstantNode ((sym->address>>8) & 0x07);	/* bit@  */
		CreateNode (sym->address & 0x8000 ? bitis0 : bitis1, 1, 0, FALSE);
		return word;

	} else {
		Error (MsgExpectedValue, 0);
		t = notype;
	}

	/*----------------------------------------------*/
	/*	String kann nochmal indiziert werden	*/
	/*----------------------------------------------*/

	if ((t != string) || (getToken() != seka)) return t;

	Consume(seka);
	GetValue(word);
	Consume(sekz);
	CreateNode (addu, 2, 0, FALSE);
	CreateLoadNode (code, 1);
	t = word;

	return t;
}


/*------------------------------------------------------*\
|*	ADDR(var) liefert Adresse der Variablen		*|
\*------------------------------------------------------*/

static Type
DoAddr(void) {
	Consume(saddr);
	Consume(srka);
	Consume(sident);
	Symbol sym = FindSymbol(getTokenName());
	(void) GetAddress(sym);
	Consume(srkz);
	return word;
}


/*------------------------------------------------------*\
|*	Textkonstante					*|
\*------------------------------------------------------*/

static uint16
DecodeStringConstant (char * buffer) {

	char * src = buffer;
	char * dst = buffer;

	if (*src=='"') src++;

	char c;
	while ((c=*src++)) {
		unsigned val = 0;
		if (c=='"') break;
		if (c=='\\') switch ((c=*src++)) {
		case 0:
			src--;
			continue;
		case 'a':
			c = '\a'; break;
		case 'b':
			c = '\b'; break;
		case 'e':
			c = '\033'; break;
		case 'f':
			c = '\f'; break;
		case 'n':
			c = '\n'; break;
		case 'r':
			c = '\r'; break;
		case 't':
			c = '\t'; break;
		case '0':
		case '1':
		case '2':
		case '3':
			val = c-'0';
			if (*src>='0' && *src<='7')
				val=(val<<3) + (*src++-'0');
			if (*src>='0' && *src<='7')
				val=(val<<3) + (*src++-'0');
			c = val;
		}
		*dst++=c;
	}
	*dst++=0;
	return CodeString(buffer, dst-buffer);
}


/*==============================================================*\
|*								*|
|*								*|
|*		Ausdrücke verschiedener Klammerebenen		*|
|*								*|
|*								*|
\*==============================================================*/

Type
DoExpression (void);


/*----------------------------------------------*\
|*						*|
|*	0: elementar oder Klammerausdruck	*|
|*						*|
\*----------------------------------------------*/

static Type
Expression0(void) {

	Type t = notype;

	switch (getToken()) {

	case sident:
		t = DoIdent();
		break;

	case saddr:
		t = DoAddr();
		break;

	case snumber:
		CreateConstantNode (getTokenValue());
		t = word;
		Cns();
		break;

	case stxt:
		Consume(stxt);
		CreateConstantNode (DecodeStringConstant(getTokenName()));
		t = string;
		break;

	case srka:
		Consume(srka);
		t = DoExpression();
		Consume(srkz);
		break;

	default:
		Fatal (MsgExpectedExpression, 0);
		break;
	}

	if (t != word) return t;

	/*------------------------------------------------------*\
	|*	Angehaengtes '^' oder '@' liest Ram oder Rom	*|
	\*------------------------------------------------------*/

	switch (getToken()) {

	case scaret:
		Consume (scaret);
		CreateLoadNode (external, 1);
		break;

	case sat:
		Consume (sat);
		CreateLoadNode (code, 1);
		break;

	default:
		break;
	}

	return t;
}


/*----------------------------------------------*\
|*						*|
|*	1: Not / Complement			*|
|*						*|
\*----------------------------------------------*/

static Type
Expression1 (void) {
	switch (getToken()) {
	case snot:
		Consume(snot);
		VerifyType (Expression1(), word);
		CreateNode (notu, 0, 0, FALSE);
		return word;
	case stilde:
		Consume(stilde);
		VerifyType (Expression1(), word);
		CreateNode (cplu, 0, 0, FALSE);
		return word;
	default: ;
	}
	return Expression0();
}


/*----------------------------------------------*\
|*						*|
|*	2: * DIV MOD				*|
|*						*|
\*----------------------------------------------*/

static Type
Expression2 (void) {

	Type t = Expression1();

	for (;;) {
		if (getToken() == smul) {
			VerifyType(t, word);
			Cns();
			t = Expression1();
			VerifyType(t, word);
			CreateNode (mulu, 2, 0, FALSE);
			continue;
		}
		if (getToken() == sdiv) {
			VerifyType(t, word);
			Cns();
			t = Expression1();
			VerifyType(t, word);
			CreateNode (divu, 2, 0, FALSE);
			continue;
		}
		if (getToken() == smod) {
			VerifyType(t, word);
			Cns();
			t = Expression1();
			VerifyType(t, word);
			CreateNode (modu, 2, 0, FALSE);
			continue;
		}
		break;
	}
	return t;
}


/*----------------------------------------------*\
|*						*|
|*	3: + - !+ !-				*|
|*						*|
\*----------------------------------------------*/

static Type
Expression3(void) {

	Type t = Expression2();

	for (;;) {
		if (getToken() == sadd) {
			VerifyType(t, word);
			Cns();
			t = Expression2();
			VerifyType(t, word);
			CreateNode (addu, 0, 2, FALSE);
			continue;
		}
		if (getToken() == saddc) {
			VerifyType(t, word);
			Cns();
			t = Expression2();
			VerifyType(t, word);
			CreateNode (adduc, 0, 2, FALSE);
			continue;
		}
		if (getToken() == ssub) {
			VerifyType(t, word);
			Cns();
			t = Expression2();
			VerifyType(t, word);
			CreateNode (subu, 0, 2, FALSE);
			continue;
		}
		if (getToken() == ssubb) {
			VerifyType(t, word);
			Cns();
			t = Expression2();
			VerifyType(t, word);
			CreateNode (subub, 0, 2, FALSE);
			continue;
		}
		break;
	}
	return t;
}


/*----------------------------------------------*\
|*						*|
|*	4: Vergleiche				*|
|*						*|
\*----------------------------------------------*/

static Type
Expression4(void) {

	Type t1 = Expression3();

	if (getToken() == seq) {
		Cns();
		Type t2 = Expression3();
		VerifyType(t2, t1);
		CreateNode (eq, 0, 1, FALSE);
		return word;
	}

	if (getToken() == sne) {
		Cns();
		Type t2 = Expression3();
		VerifyType(t2, t1);
		CreateNode (ne, 0, 1, FALSE);
		return word;
	}

	if (getToken() == slt) {
		Cns();
		Type t2 = Expression3();
		VerifyType(t2, t1);
		CreateNode (ltu, 0, 1, FALSE);
		return word;
	}

	if (getToken() == sge) {
		Cns();
		Type t2 = Expression3();
		VerifyType(t2, t1);
		CreateNode (geu, 0, 1, FALSE);
		return word;
	}

	if (getToken() == sgt) {
		Cns();
		Type t2 = Expression3();
		VerifyType(t2, t1);
		CreateNode (ltu, 0, 1, TRUE);
		return word;
	}

	if (getToken() == sle) {
		Cns();
		Type t2 = Expression3();
		VerifyType(t2, t1);
		CreateNode (geu, 0, 1, TRUE);
		return word;
	}

	if (getToken() == sin) {
		Cns();
		Type t2 = Expression3();
		VerifyType(t2, t1);
		CreateNode (bitis1, 1, 1, TRUE);
		return word;
	}

	return t1;
}


/*----------------------------------------------*\
|*						*|
|*	5: And					*|
|*						*|
\*----------------------------------------------*/

static Type
Expression5 (void) {

	Type t = Expression4();
	while (getToken() == sand) {
		VerifyType(t, word);
		Consume(sand);
		t = Expression4();
		VerifyType(t, word);
		CreateNode (andu, 0xff, 0, FALSE);
	}
	return t;
}


/*----------------------------------------------*\
|*						*|
|*	6: Or					*|
|*						*|
\*----------------------------------------------*/

static Type
Expression6 (void) {

	Type t = Expression5();

	for (;;) {
		if(getToken() == sor) {
			VerifyType(t, word);
			Cns();
			t = Expression5();
			VerifyType(t, word);
			CreateNode (oru, 0, 0, FALSE);
			continue;
		}
		if (getToken() == sxor) {
			VerifyType(t, word);
			Cns();
			t = Expression5();
			VerifyType(t, word);
			CreateNode (xoru, 0, 0, FALSE);
			continue;
		}
		break;
	}
	return t;
}


/*==============================================================*\
|*								*|
|*								*|
|*	Von aussen zugreifbare Routinen				*|
|*								*|
|*								*|
\*==============================================================*/


/*------------------------------------------------------*\
|*	Parse eine Expression,				*|
|*	Liefere Wert im Baum und Typ als Ergebnis	*|
\*------------------------------------------------------*/

Type
DoExpression (void) {
	return Expression6();
}


/*------------------------------------------------------*\
|*	Parse eine Expression, erwarte bestimmten Typ	*|
|*	Liefere Zeiger auf die Expression,		*|
|*	*ohne* die Expression zu "verbrauchen".		*|
\*------------------------------------------------------*/

Node
GetValue (Type requested_type) {
	if (requested_type==byte) requested_type=word;
	VerifyType (DoExpression(), requested_type);
	return getLastInserted();
}


/*------------------------------------------------------*\
|*	Parse Expression, erwarte bestimmten Typ.	*|
|*	Erwarte eine Konstante und liefere den Wert.	*|
\*------------------------------------------------------*/

uint16
ConstantValueOfType (Type type) {
	GetValue (type!=byte? type : byte);
	CreateNode (svoid, typesize(type), 0, FALSE);
	Node node = getLastUnused();
	OptimizeNodes();
	if (node->right->op == cval) return node->right->value;
	Error (MsgExpectedConstant, 0);
	return 0;
}


/*------------------------------------------------------*\
|*	Parse konstante Expression.			*|
|*	Liefere Typ als Parameter und Wert als Ergebnis	*|
\*------------------------------------------------------*/

uint16
ConstantValue (Type* typep) {
	Type type = DoExpression();
	CreateNode (svoid, typesize(type), 0, FALSE);
	Node node = getLastUnused();
	OptimizeNodes();
	*typep = type;
	if (node->right->op == cval) return node->right->value;
	Error (MsgExpectedConstant, 0);
	return 0;
}


/*===================================================================*\
|*	$Id: expressions.c,v 1.2 2019/08/03 00:05:00 wolf Exp $
\*===================================================================*/
