/****************************************************************)
(*								*)
(*								*)
(*	N       N        AAA       N       N      OOOOO		*)
(*	NN      N       A   A      NN      N     O     O	*)
(*	N N     N      A     A     N N     N    O       O	*)
(*	N  N    N     A       A    N  N    N    O       O	*)
(*	N   N   N     A       A    N   N   N    O       O	*)
(*	N    N  N     AAAAAAAAA    N    N  N    O       O	*)
(*	N     N N     A       A    N     N N    O       O	*)
(*	N      NN     A       A    N      NN     O     O	*)
(*	N       N     A       A    N       N      OOOOO		*)
(*								*)
(*								*)
(****************************************************************)
(*	Compiler fuer ein *sehr* eingeschraenktes Pascal.	*)
(*	Erzeugt Maschinencode fuer 8031 & co Prozessoren.	*)
(****************************************************************/

/*===================================================================*\
|*	$Id: memory.c,v 1.1 2018/12/21 00:14:50 wolf Exp $
|*===================================================================*|
|*
|*	$Log: memory.c,v $
|*	Revision 1.1  2018/12/21 00:14:50  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 22:27:22  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 16:03:56  wolf
|*	Initial revision
|*
|*	Revision 1.3  2011/08/09 17:26:17  wolf
|*	* Cleanup
|*
|*	Revision 1.2  2008/12/12 20:22:21  wolf
|*	*** empty log message ***
|*
|*===================================================================*|
|*	1988-1996 als Turbopascal-Programm für 8080 und 8086
\*===================================================================*/

#include "nano.h"

#include "memory.h"
#include "symboltable.h"

#include <stdio.h>
#include <string.h>


uint16	Cadrs = 0x0040;
uint16	Cadre = 0x1fff;
uint16	Dadrs = 0x8000;
uint16	Dadre = 0xffff;
uint16	Ladrs =   0x20;
uint16	Ladre =   0x7f;
uint16	Uadrs =   0x80;
uint16	Uadre =   0xbf;

uint16	Cadr  = 0x0040;
uint16	Dadr  = 0x8000;
uint16	Ladr  =   0x20;
uint16	Uadr  =   0xc0;
uint16	Tadr  = 0x2000;

void
MemoryReset (void) {
	Cadr = Cadrs;
	Tadr = Cadre + 1;
	Dadr = Dadrs;
	Ladr = Ladrs;
	Uadr = Uadre + 1;
}

uint16
AllocateIntern (uint16 size) {
	uint16 result = Ladr;
	Ladr += size;
	if (Ladr > Ladre || Ladr > 0x80 || Ladr > Uadr)
		Error (MsgResource, "internal memory");
	return result;
}

uint16
AllocateInternUpper (uint16 size) {
	Uadr -= size;
	if (Uadr<Ladr)
		Error (MsgResource, "internal memory");
	return Uadr;
}

uint16
AllocateExtern (uint16 size) {
	if (Dadr < Dadrs || Dadr > Dadre)
		Error (MsgResource, "external memory");
	if (Dadr+size-1 < Dadrs || Dadr+size-1 > Dadre)
		Error (MsgResource, "external memory");
	uint16 result = Dadr;
	Dadr += size;
	return result;
}


/****************************************/
/*                                      */
/*	Allocate variable		*/
/*                                      */
/****************************************/

void
AllocateMemory (Symbol sym) {

	/*------------------------------------------------------*\
	|*	Absolute Variablen nicht allozieren		*|
	\*------------------------------------------------------*/

	if (sym->absolute) {
		if (sym->storage != external && !sym->size && sym->address >= 0x80)
			sym->storage = reg;
		if (sym->storage == upb)
			sym->storage = internal;
		return;
	}

	/*------------------------------------------------------*\
	|*	Nicht absolut ... bestimme Platzbedarf		*|
	\*------------------------------------------------------*/

	short size = 2;

	if (sym->indirect) {
		if (!sym->size) size = 1;
	} else {
		if (sym->type==byte) size  = 1;
		if (sym->size!=   0) size *= sym->size;
	}

	/*------------------------------------------------------*\
	|*	Alloziere internen Speicher von oben		*|
	\*------------------------------------------------------*/

	if (sym->storage == upb) {
		sym->address = AllocateInternUpper(size);
		sym->storage = internal;
		return;
	}

	/*------------------------------------------------------*\
	|*	Alloziere internen Speicher von unten		*|
	\*------------------------------------------------------*/

	if (sym->indirect || (sym->storage == internal)) {
		sym->address = AllocateIntern(size);
		return;
	}

	/*------------------------------------------------------*\
	|*	Alloziere externen Speicher			*|
	\*------------------------------------------------------*/

	if (sym->storage == external) {
		sym->address = AllocateExtern(size);
	}
}


/*===================================================================*\
|*	$Id: memory.c,v 1.1 2018/12/21 00:14:50 wolf Exp $
\*===================================================================*/
