/****************************************************************)
(*								*)
(*								*)
(*	N       N        AAA       N       N      OOOOO		*)
(*	NN      N       A   A      NN      N     O     O	*)
(*	N N     N      A     A     N N     N    O       O	*)
(*	N  N    N     A       A    N  N    N    O       O	*)
(*	N   N   N     A       A    N   N   N    O       O	*)
(*	N    N  N     AAAAAAAAA    N    N  N    O       O	*)
(*	N     N N     A       A    N     N N    O       O	*)
(*	N      NN     A       A    N      NN     O     O	*)
(*	N       N     A       A    N       N      OOOOO		*)
(*								*)
(*								*)
(****************************************************************)
(*	Compiler fuer ein *sehr* eingeschraenktes Pascal.	*)
(*	Erzeugt Maschinencode fuer 8031 & co Prozessoren.	*)
(****************************************************************/

/*===================================================================*\
|*	$Id: preprocessor.c,v 1.1 2018/12/21 00:15:19 wolf Exp $
|*===================================================================*|
|*
|*	$Log: preprocessor.c,v $
|*	Revision 1.1  2018/12/21 00:15:19  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 22:27:22  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 16:03:56  wolf
|*	Initial revision
|*
|*	Revision 1.4  2011/08/09 17:26:17  wolf
|*	* Cleanup
|*
|*	Revision 1.3  2010/12/24 16:26:09  wolf
|*	*** empty log message ***
|*
|*	Revision 1.2  2008/12/12 20:22:21  wolf
|*	*** empty log message ***
|*
|*	Revision 1.1  2008/12/07 09:43:42  wolf
|*	Initial revision
|*
|*===================================================================*|
|*	1988-1996 als Turbopascal-Programm für 8080 und 8086
\*===================================================================*/


#include "nano.h"
#include "preprocessor.h"

#include <string.h>
#include <strings.h>
#include <stdlib.h>


/*==============================================================*\
|*								*|
|*	Datenstruktur 				 		*|
|*								*|
\*==============================================================*/


static struct {
	FILE *		file;
	const char *	filename;
	unsigned	linenumber;
} Files[MaxIncludeLevel];

static int
CurrentIncludeLevel = -1;

/*--------------------------------------------------------------*\
|*	Zugriff auf die Datenstruktur fuer Fehleranzeige	*|
\*--------------------------------------------------------------*/

const char *
getCurrentFilename (void) {
	if (CurrentIncludeLevel>=0 && CurrentIncludeLevel<MaxIncludeLevel)
		return Files[CurrentIncludeLevel].filename;
	return "-";
}

unsigned
getCurrentLinenumber (void) {
	if (CurrentIncludeLevel>=0 && CurrentIncludeLevel<MaxIncludeLevel)
		return Files[CurrentIncludeLevel].linenumber;
	return 0;
}


/*==============================================================*\
|*								*|
|*	Oeffnen der Hauptdatei oder einer Include-Datei		*|
|*								*|
\*==============================================================*/

BOOL
include (const char filename[]) {

	if (!filename) return FALSE;

	/*------------------------------------------------------*\
	|*	Ueberpruefe die Schachteltiefe			*|
	\*------------------------------------------------------*/

	int newlevel=CurrentIncludeLevel+1;

	if (newlevel >= MaxIncludeLevel) {
		Error (MsgLimit, "MaxIncludeLevel");
		return FALSE;
	}

	/*------------------------------------------------------*\
	|*	Bestimme den Dateinamen				*|
	\*------------------------------------------------------*/

	if (filename[0] != '/' && CurrentIncludeLevel>=0 && strchr(Files[CurrentIncludeLevel].filename,'/')) {
		/*------------------------------------------------------*\
		|*	wenn includefile, relativ zur ladenden Datei	*|
		\*------------------------------------------------------*/

		char * buffer = (char*) malloc (strlen(Files[CurrentIncludeLevel].filename)
			+ strlen (filename) + 3);
		strcpy (buffer, Files[CurrentIncludeLevel].filename);
		strcpy (strrchr (buffer, '/')+1, filename);
		Files[newlevel].filename = buffer;

	} else {

		/*------------------------------------------------------*\
		|*	wenn Hauptdatei, relativ zum Arbeitsverzeichnis	*|
		\*------------------------------------------------------*/

		char * copy = (char*) malloc (strlen(filename)+1);
		strcpy (copy, filename);
		Files[newlevel].filename = copy;
	}

	/*------------------------------------------------------*\
	|*	Oeffne die Datei				*|
	\*------------------------------------------------------*/

	Files[newlevel].file      = fopen (Files[newlevel].filename, "r");
	Files[newlevel].linenumber= 0;

	if (!Files[newlevel].file) {
		Error (MsgFopen, Files[newlevel].filename);
		return FALSE;
	}

	/*------------------------------------------------------*\
	|*	Ok, aktiviere die neue Schachtelebene		*|
	\*------------------------------------------------------*/

	CurrentIncludeLevel = newlevel;
	return TRUE;
}


/*==============================================================*\
|*								*|
|*	Liefere eine Zeile der jeweiligen Datei			*|
|*	Bei Ende einer Includedatei wird einmalig ein leerer	*|
|*	String (Laenge=0) als Endemarker geliefert.		*|
|*								*|
\*==============================================================*/

static char Line [MaxInputLineLength];

extern BOOL
isDefined (const char symbol[]);

const char *
getLine (void) {

	unsigned Condition=0;

	while (CurrentIncludeLevel>=0 && CurrentIncludeLevel < MaxIncludeLevel) {

		/*------------------------------------------------------*\
		|*	Lese eine Zeile					*|
		\*------------------------------------------------------*/

		++Files[CurrentIncludeLevel].linenumber;

		/*------------------------------------------------------*\
		|*	Bei Lesefehler oder Ende der Datei:		*|
		|*	- schließe die Datei				*|
		|*	- verlasse die Schachtelebene			*|
		|*	- liefere einen leeren String als Ende-Marker	*|
		\*------------------------------------------------------*/

		if (!fgets (Line, MaxInputLineLength, Files[CurrentIncludeLevel].file)) {
			if (ferror (Files[CurrentIncludeLevel].file))
				Error (MsgFread, Files[CurrentIncludeLevel].filename);
			fclose (Files[CurrentIncludeLevel].file);
			CurrentIncludeLevel--;
			Condition = 0;
			return "";
		}

		/*------------------------------------------------------*\
		|*	Schreibe Zeile ins Listingfile,			*|
		|*	wenn Condition vor oder nach Zeile es erlaubt	*|
		\*------------------------------------------------------*/

		int lst = !Condition;
		if (!strncmp(Line, ".eof"  , 4)) lst = 1;
		if (!strncmp(Line, ".else" , 5)) lst = !(Condition>>1);
		if (!strncmp(Line, ".endif", 6)) lst = !(Condition>>1);

		if (lst && chkOption ('l', 0)) {
			lstfileselect ("line");
			fprintf (lstfile, "; %s:%u: %s",
				Files[CurrentIncludeLevel].filename,
				Files[CurrentIncludeLevel].linenumber,
				Line);
		}

		/*------------------------------------------------------*\
		|*	Wenn die bedingte Compilierung es erlaubt	*|
		|*	(Condition==0), liefere eine normale Zeile aus	*|
		\*------------------------------------------------------*/

		if (Line[0]!='.') {

			if (Condition) continue;
			return Line;
		}

		/*------------------------------------------------------*\
		|*	Zeilen, die mit einem Punkt "." beginnen,	*|
		|*	sind Preprozessor-Befehle			*|
		|*------------------------------------------------------*|
		|*	Ja, ich *weiss*, dass das "#" das geeignete	*|
		|*	Zeichen ist, aber das Nanopascal ist zu einer	*|
		|*	Teit entstanden, als ich viel mit WordStar	*|
		|*	und noch nicht mit C gearbeitet habe :-)	*|
		\*------------------------------------------------------*/

		const char* cmd = strtok (&Line[1], " \t\r\n");
		if (!cmd) cmd="";

		/*------------------------------------------------------*\
		|*	Diese Kommandos werden immer ausgeführt,	*|
		|*	auch wenn Condition != 0 ist.			*|
		\*------------------------------------------------------*/

		if (!strcasecmp (cmd, "eof")) {
			fclose (Files[CurrentIncludeLevel].file);
			CurrentIncludeLevel--;
			Condition = 0;
			return "";
		}

		if (!strcasecmp (cmd, "ifdef")) {
			Condition <<= 1;
			if (!isDefined (strtok (NULL, " \t\t\n")))
				Condition|=1;
			continue;
		}

		if (!strcasecmp (cmd, "ifndef")) {
			Condition <<= 1;
			if (isDefined (strtok (NULL, " \t\t\n")))
				Condition|=1;
			continue;
		}

		if (!strcasecmp (cmd, "else")) {
			Condition ^= 1;
			continue;
		}

		if (!strcasecmp (cmd, "endif")) {
			Condition >>= 1;
			continue;
		}

		/*------------------------------------------------------*\
		|*	Die weiteren Kommandos werden nur ausgeführt,	*|
		|*	wenn die bedingte Compilierung dies erlaubt.	*|
		|*	(Condition==0)					*|
		\*------------------------------------------------------*/

		if (Condition) continue;

		/*--------------------------------------*\
		|*	Preprozessorkommandos		*|
		\*--------------------------------------*/


		if (!strcasecmp (cmd, "show")) {
			FILE*file = restrictedMode? errfile : stderr;
			fprintf (file, "*** %s\n", strtok (NULL, "\r\n"));
			continue;
		}

		if (!strcasecmp (cmd, "fi")) {
			if (restrictedMode) {
				Fatal (MsgRestrictedMode, "Datei-Include micht erlaubt.");
			}
			include (strtok (NULL, " \t\r\n"));
			continue;
		}
		
		if (!strcasecmp (cmd, "assert")) {
			const char * text;
			extern void AssertRef (uint16 data);
			while ((text = strtok (NULL, " \t\r\n"))) {
				AssertRef (strtol (text, NULL, 16));
			}
			continue;
		}
		
		if (strlen(cmd)==2 && cmd[1]=='+') {
			SetOption (cmd[0],1);
			continue;
		}
		
		if (strlen(cmd)==2 && cmd[1]=='-') {
			ClrOption (cmd[0]);
			continue;
		}
		
		if (strlen(cmd)>=3 && cmd[1]=='=') {
			SetOption (cmd[0], atoi(&cmd[2]));
			continue;
		}
		
		Error (MsgPrepUnknown, cmd);
	}

	Error (MsgPrepEOF, 0);

	return NULL;
}

/*===================================================================*\
|*	$Id: preprocessor.c,v 1.1 2018/12/21 00:15:19 wolf Exp $
\*===================================================================*/
