/****************************************************************)
(*								*)
(*								*)
(*	N       N        AAA       N       N      OOOOO		*)
(*	NN      N       A   A      NN      N     O     O	*)
(*	N N     N      A     A     N N     N    O       O	*)
(*	N  N    N     A       A    N  N    N    O       O	*)
(*	N   N   N     A       A    N   N   N    O       O	*)
(*	N    N  N     AAAAAAAAA    N    N  N    O       O	*)
(*	N     N N     A       A    N     N N    O       O	*)
(*	N      NN     A       A    N      NN     O     O	*)
(*	N       N     A       A    N       N      OOOOO		*)
(*								*)
(*								*)
(****************************************************************)
(*	Compiler fuer ein *sehr* eingeschraenktes Pascal.	*)
(*	Erzeugt Maschinencode fuer 8031 & co Prozessoren.	*)
(****************************************************************/

/*===================================================================*\
|*	$Id: scanner.c,v 1.1 2018/12/21 00:15:19 wolf Exp $
|*===================================================================*|
|*
|*	$Log: scanner.c,v $
|*	Revision 1.1  2018/12/21 00:15:19  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 22:27:22  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 16:03:56  wolf
|*	Initial revision
|*
|*	Revision 1.4  2011/08/09 17:26:17  wolf
|*	* Cleanup
|*
|*	Revision 1.3  2010/12/24 16:26:09  wolf
|*	*** empty log message ***
|*
|*	Revision 1.2  2008/12/12 20:22:21  wolf
|*	*** empty log message ***
|*
|*===================================================================*|
|*	1988-1996 als Turbopascal-Programm für 8080 und 8086
\*===================================================================*/

#include "nano.h"

#include "scanner.h"
#include "preprocessor.h"

#include <string.h>
#include <strings.h>
#include <ctype.h>
#include <stdio.h>


/*==============================================================*\
|*								*|
|*	Datenstruktur 				 		*|
|*								*|
\*==============================================================*/

static const char Empty[] = "";
static const char* Line  = Empty;
static const char* Next  = Empty;
static const char* TokenPos= NULL;

/*--------------------------------------------------------------*\
|*	Zugriff auf die Datenstruktur fuer Fehleranzeige	*|
\*--------------------------------------------------------------*/

unsigned
getCurrentColumn (void) {
	return TokenPos? TokenPos - Line : 0;
}

/*--------------------------------------------------------------*\
|*	Das letzte gescannte Token				*|
\*--------------------------------------------------------------*/

Token		TokenType	= snul;
unsigned	TokenValue	= 0;
char		TokenName[MaxNameLength];

/*--------------------------------------------------------------*\
|*	Zugriff auf die Tokendaten fuer den Parser		*|
\*--------------------------------------------------------------*/

char *
getTokenName (void) { return TokenName; }

uint16
getTokenValue (void) { return TokenValue; }


/*==============================================================*\
|*								*|
|*	Keywordtabelle						*|
|*								*|
\*==============================================================*/


static struct keyword {
	const char*	name;
	Token		value;
} Keywords[] = {
	{ "ADDR",	saddr		},
	{ "AND",	sand		},
	{ "ARRAY",	sarray		},
	{ "BEGIN",	sbegin		},
	{ "BREAK",	sbreak		},
	{ "BYTE",	sbyte		},
	{ "CHAR",	sbyte		},
	{ "CASE",	scase		},
	{ "CLR",	sclr		},
	{ "CONST",	sconst		},
	{ "CONTINUE",	scontinue	},
	{ "CPL",	scpl		},
	{ "DECR",	sdecr		},
	{ "DIV",	sdiv		},
	{ "DO",		sdo		},
	{ "ELSE",	selse		},
	{ "END",	send		},
	{ "EXIT",	sexit		},
	{ "EXTERN",	sextern		},
	{ "FOR",	sfor		},
	{ "HI",		shi		},
	{ "IF",		sif		},
	{ "IN",		sin		},
	{ "INCR",	sincr		},
	{ "INLINE",	sinline		},
	{ "INTEGER",	sword		},
	{ "INTERN",	sintern		},
	{ "INTERRUPT",	sinterrupt	},
	{ "LO",		slo		},
	{ "LOCK",	slock		},
	{ "LOOP",	sloop		},
	{ "MOD",	smod		},
	{ "NOT",	snot		},
	{ "OF",		sof		},
	{ "OR",		sor		},
	{ "PROCEDURE",	sprocedure	},
	{ "PROGRAM",	sprogram	},
	{ "REPEAT",	srepeat		},
	{ "RETURN",	sreturn		},
	{ "QUICK",	squick		},
	{ "SET",	sset		},
	{ "SHL",	sshl		},
	{ "SHR",	sshr		},
	{ "STRING",	sstring		},
	{ "TABLE",	stable		},
	{ "THEN",	sthen		},
	{ "TO",		sto		},
	{ "UNTIL",	suntil		},
	{ "VAR",	svar		},
	{ "WHILE",	swhile		},
	{ "WORD",	sword		},
	{ "XOR",	sxor		},
	{ "",		snul		},
/*----------------------------------------------*\
|*	Leerstring markiert Ende der Keywords	*|
\*----------------------------------------------*/
	{ "!",		sexclam		},
	{ "!+",		saddc		},
	{ "!-",		ssubb		},
	{ "(",		srka		},
	{ ")",		srkz		},
	{ "*",		smul		},
	{ "+",		sadd		},
	{ ",",		scomma		},
	{ "-",		ssub		},
	{ ".",		sdot		},
	{ "..",		supto		},
	{ ":",		scolon		},
	{ ":=",		sassign		},
	{ ";",		ssemi		},
	{ ";",		ssemi		},
	{ "<",		slt		},
	{ "<=",		sle		},
	{ "<>",		sne		},
	{ "=",		seq		},
	{ ">",		sgt		},
	{ ">=",		sge		},
	{ "@",		sat		},
	{ "[",		seka		},
	{ "]",		sekz		},
	{ "^",		scaret		},
	{ "~",		stilde		},
	{ "<identifier>", sident	},
	{ "<eof>",	seof		},
	{ NULL,		snul		},
};


/*==============================================================*\
|*								*|
|*	Die Basisscanroutine			 		*|
|*								*|
\*==============================================================*/

static void
Scan (void) {

	/*--------------------------------------*\
	|*	Initialisiere das Token		*|
	\*--------------------------------------*/

	TokenPos  = NULL;
	TokenType = snul;
	TokenName[0] = 0;
	TokenValue= 0;

	/*--------------------------------------*\
	|*	Line==NULL => end of files	*|
	\*--------------------------------------*/

	while (Line) {

		/*--------------------------------------*\
		|*	Ueberspringe Blanks		*|
		|*	Bei EOL: hole naechste Zeile	*|
		\*--------------------------------------*/

		char c, ch=*Next;

		while (ch && isspace(ch)) ch=*++Next;

		if (!ch) {
			Next = Line = getLine();
			continue;
		}

		/*--------------------------------------*\
		|*	Merke Token-Startposition	*|
		\*--------------------------------------*/

		TokenPos = Next++;

		switch (ch) {

		case '!':
			if (*Next=='+') {
				Next++;
				TokenType = saddc;
				break;
			}
			if (*Next=='-') {
				Next++;
				TokenType = ssubb;
				break;
			}
				TokenType = sexclam;
			break;

		case '"':
			/*--------------------------------------*\
			|*	Stringkonstante			*|
			|*	Wird durch '"' oder EOL beendet	*|
			\*--------------------------------------*/

			while ((Next = strchr (Next, '"'))) {
				Next++;
				if (Next[-2]!='\\') break;
			}
			if (!Next) {
				Error (MsgScanString, 0);
				Next = strchr (TokenPos, 0);
			}
			TokenType = stxt;
			break;

		case '%':
			/*--------------------------------------*\
			|*	Binaerkonstante.		*|
			|*	Inspiriert durch Assembler	*|
			\*--------------------------------------*/

			for (c=*Next; c=='0' || c=='1'; c=*++Next)
				TokenValue = TokenValue*2 + (c-'0');
			if (Next==TokenPos+1) break;
			TokenType = snumber;
			break;

		case '$':
			/*--------------------------------------*\
			|*	Hexadezimalkonstante.		*|
			|*	Inspiriert durch Assembler	*|
			\*--------------------------------------*/

			for (c=*Next; isxdigit(c); c=*++Next) {
				if (c>='@') c+= 9;
				TokenValue = TokenValue*16 + (c&0x0f);
			}
			if (Next==TokenPos+1) break;
			TokenType = snumber;
			break;

		case '&':
			TokenType = sand;
			break;

		case '(':
			TokenType = srka;
			break;

		case ')':
			TokenType = srkz;
			break;

		case '*':
			TokenType = smul;
			break;

		case '+':
			TokenType = sadd;
			break;

		case ',':
			TokenType = scomma;
			break;

		case '-':
			TokenType = ssub;
			break;

		case '.':
			if (*Next=='.') {
				Next++;
				TokenType = supto;
				break;
			}
			TokenType = sdot;
			break;

		case '0':
			/*--------------------------------------*\
			|*	Binaerkonstante.		*|
			\*--------------------------------------*/

			if (Next[0]=='b' && (Next[1]=='0'|| Next[1]=='1')) {
				char c;
				for (c=*++Next; c=='0' || c=='1'; c=*++Next)
					TokenValue = TokenValue*2 + (c-'0');
				TokenType = snumber;
				break;
			}

			/*--------------------------------------*\
			|*	Hexadezimalkonstante.		*|
			|*	Inspiriert durch C		*|
			\*--------------------------------------*/

			if (Next[0]=='x' && isxdigit(Next[1])) {

				for (c=*++Next; isxdigit(c); c=*++Next) {
					if (c>='@') c+= 9;
					TokenValue = TokenValue*16 + (c&0x0f);
				}
				TokenType = snumber;
				break;
			}
			/* fall through */

		case '1': case '2': case '3':
		case '4': case '5': case '6':
		case '7': case '8': case '9':

			TokenValue = ch - '0';
			for (c=*Next; isdigit(c); c=*++Next)
				TokenValue = TokenValue*10 + (c-'0');
			TokenType = snumber;
			break;

		case ':':
			if (*Next=='=') {
				Next++;
				TokenType = sassign;
				break;
			}
			TokenType = scolon;
			break;

		case ';':
			TokenType = ssemi;
			break;

		case '<':
			if (*Next=='>') {
				Next++;
				TokenType = sne;
				break;
			}
			if (*Next=='=') {
				Next++;
				TokenType = sle;
				break;
			}
			TokenType = slt;
			break;

		case '=':
			TokenType = seq;
			break;

		case '>':
			if (*Next=='=') {
				Next++;
				TokenType = sge;
				break;
			}
			TokenType = sgt;
			break;

		case '@':
			TokenType = sat;
			break;

		case '[':
			TokenType = seka;
			break;

		case ']':
			TokenType = sekz;
			break;

		case '^':
			TokenType = scaret;
			break;

		case '{':
			while (Line && *Line) {
				char*eoc = strchr(Next, '}');
				if (eoc) {
					Next = eoc+1;
					break;
				}
				Next = Line = getLine();
			}
			continue;

		case '|':
			TokenType = sor;
			break;

			break;

		case '\'':
			if (!Next[0]) break;
			if (Next[1]=='\'') {
				TokenValue = Next[0];
				Next+=2;
				TokenType = snumber;
				break;
			}
			if ((Next[0]&0xe0)==0xc0 && (Next[1]&0xc0)==0x80 && Next[2]=='\'') {
				TokenValue = ((Next[0]&0x1f)<<6) | (Next[1]&0x3f);
				Next+=3;
				TokenType = snumber;
				break;
			}
			if ((Next[0]&0xf0)==0xe0 && (Next[1]&0xc0)==0x80 && (Next[2]&0xc0)==0x80 && Next[3]=='\'') {
				TokenValue = ((Next[0]&0x0f)<<12) | ((Next[1]&0x3f)<<6) | (Next[2]&0x3f);
				Next+=4;
				TokenType = snumber;
				break;
			}
			break;

		case '~':
			TokenType = stilde;
			break;

		default:

			if (!isalpha(ch)) break;

			/*--------------------------------------*\
			|*	Identifier			*|
			\*--------------------------------------*/

			for (c=*Next; isalnum(c)||c=='?'||c=='_'; c=*++Next);
			TokenType = sident;
		}

		/*----------------------------------------------*\
		|*	Kein Token gefunden			*|
		|*	Verlängere Zeichen (wegen utf8 :-)	*|
		\*----------------------------------------------*/

		if (TokenType==snul) {
			while (*Next&0x80) Next++;
		}

		/*----------------------------------------------*\
		|*	Sichere den Tokentext nach TokenName	*|
		\*----------------------------------------------*/

		if (Next-TokenPos < MaxNameLength) {
			strncpy (TokenName, TokenPos, Next-TokenPos);
			TokenName[Next-TokenPos] = 0;
		} else {
			strncpy (TokenName, TokenPos, MaxNameLength-1);
			TokenName[MaxNameLength-1] = 0;
			Error (MsgLimit, "MaxNameLength");
		}

		/*----------------------------------------------*\
		|*	Kein Token gefunden			*|
		\*----------------------------------------------*/

		if (TokenType==snul) {
			Error (MsgScanUnknown, TokenName);
			continue;
		}

		/*----------------------------------------------*\
		|*	Wenn Identifier, suche in Keywords	*|
		\*----------------------------------------------*/

		if (TokenType == sident) {

			struct keyword * kp;
			for (kp=Keywords; *kp->name; kp++)
				if (!strcasecmp (TokenName, kp->name))
					TokenType = kp->value;
		}

		/*----------------------------------------------*\
		|*	Debug: Anzeige der Token		*|
		\*----------------------------------------------*/

		if (chkOption('l','t')) {
			lstfileselect ("token");
			fprintf (lstfile, "\t<%s>\n", TokenName);
		}

		return;

	}

	TokenType = seof;
}


/*==============================================================*\
|*								*|
|*	Zugriffsroutinen fuer den Parser			*|
|*								*|
\*==============================================================*/


/*------------------------------------------------------*\
|*	Hole aktuelles Token (scanne wenn noetig)	*|
\*------------------------------------------------------*/

Token
getToken (void) {
	if (TokenType==snul) Scan();
	return TokenType;
}


/*--------------------------------------------------------------*\
|*	Verbrauche das aktuelle Token (scanne wenn noetig)	*|
\*--------------------------------------------------------------*/

void
Cns (void) {
	if (TokenType==snul) Scan();
	TokenType=snul;
}

/*--------------------------------------------------------------*\
|*	Verbrauche ein bestimmtes Token (scanne wenn noetig)	*|
\*--------------------------------------------------------------*/

void
Consume (Token token) {

	if (TokenType==snul) Scan();

	if (TokenType==token) {
		TokenType=snul;
		return;
	}

	const char * name = "token";

	for (struct keyword *kp=Keywords; kp->name; kp++)
		if (kp->value==token) name = kp->name;

	Fatal (MsgExpectedToken, name);
}


/*===================================================================*\
|*	$Id: scanner.c,v 1.1 2018/12/21 00:15:19 wolf Exp $
\*===================================================================*/
