/****************************************************************)
(*								*)
(*								*)
(*	N       N        AAA       N       N      OOOOO		*)
(*	NN      N       A   A      NN      N     O     O	*)
(*	N N     N      A     A     N N     N    O       O	*)
(*	N  N    N     A       A    N  N    N    O       O	*)
(*	N   N   N     A       A    N   N   N    O       O	*)
(*	N    N  N     AAAAAAAAA    N    N  N    O       O	*)
(*	N     N N     A       A    N     N N    O       O	*)
(*	N      NN     A       A    N      NN     O     O	*)
(*	N       N     A       A    N       N      OOOOO		*)
(*								*)
(*								*)
(****************************************************************)
(*	Compiler fuer ein *sehr* eingeschraenktes Pascal.	*)
(*	Erzeugt Maschinencode fuer 8031 & co Prozessoren.	*)
(****************************************************************/

/*===================================================================*\
|*	$Id: statements.c,v 1.1 2018/12/21 00:15:19 wolf Exp $
|*===================================================================*|
|*
|*	$Log: statements.c,v $
|*	Revision 1.1  2018/12/21 00:15:19  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 22:27:22  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 16:03:56  wolf
|*	Initial revision
|*
|*	Revision 1.5  2011/08/09 17:26:17  wolf
|*	* Cleanup
|*
|*	Revision 1.4  2010/12/24 16:26:09  wolf
|*	*** empty log message ***
|*
|*	Revision 1.3  2008/12/12 20:22:21  wolf
|*	*** empty log message ***
|*
|*===================================================================*|
|*	1988-1996 als Turbopascal-Programm für 8080 und 8086
\*===================================================================*/


#include "nano.h"

#include "tree.h"
#include "symboltable.h"
#include "scanner.h"
#include "expressions.h"
#include "statements.h"
#include "code.h"
#include "compile.h"

#include <stdio.h>



Scope	ThisScope;

/*==============================================================*\
|*								*|
|*								*|
|*		Vorspann und Nachspann für Prozeduren		*|
|*								*|
|*								*|
\*==============================================================*/


/*----------------------------------------------*\
|*						*|
|*	Code bei Betreten eines Scopes		*|
|*						*|
\*----------------------------------------------*/

void
DoEnter (int options) {

	switch (ThisScope) {

	case program:
		break;

	case procedure:
		break;

	case interrupt:
		Code2(0xc0, PSW);
		Code2(0xc0, ACC);
		Code2(0xc0, DPL);
		Code2(0xc0, DPH);
		Code2(0xc0, B);
		if (options) Code3(0x75, 0xd0, options);
		break;

	case fastinterrupt:
		break;
	}
}


/*----------------------------------------------*\
|*						*|
|*	Code bei Verlassen eines Scopes		*|
|*						*|
\*----------------------------------------------*/

void
DoLeave (void) {
	switch (ThisScope) {

	case program:
		/* Power down */
		Code3(0x43, 0x87, 0x03);
		Code2(0x80, 0xfb);
		break;

	case procedure:
		Code1(0x22);	/* RET */
		break;

	case interrupt:
		Code2(0xd0, B);
		Code2(0xd0, DPH);
		Code2(0xd0, DPL);
		Code2(0xd0, ACC);
		Code2(0xd0, PSW);

	case fastinterrupt:
		Code1(0x32);	/* RETI */
		break;
	}
}


/*----------------------------------------------*\
|*						*|
|*	Code bei Nichtbetreten eines Scopes	*|
|*						*|
\*----------------------------------------------*/

void
DoLeaveImmediately (void) {

	switch (ThisScope) {

	case program:
		/* Power down */
		Code3(0x43, 0x87, 0x03);
		Code2(0x80, 0xfb);
		break;

	case procedure:
		Code1(0x22);
		break;

	case interrupt:
	case fastinterrupt:
		Code1(0x32);
		break;
	}
}


/*==============================================================*\
|*								*|
|*								*|
|*		Elementare ungeschachtelte Befehle		*|
|*								*|
|*								*|
\*==============================================================*/


/*----------------------------------------------*\
|*						*|
|*	Proceduraufruf oder Zuweisung		*|
|*						*|
\*----------------------------------------------*/

static void
DoAssignOrCall(void) {

	Consume (sident);
	Symbol sym = FindSymbol(getTokenName());

	/*----------------------*/
	/*	Procedure Call	*/
	/*----------------------*/

	if (sym->type == proc) {

		unsigned i;
		for (i=1; i<=sym->size; i++) {

			/*----------------------*/
			/*	Parameter	*/
			/*----------------------*/

			Symbol parameter = sym + i;

			Consume(i==1 ? srka : scomma);

			CreateConstantNode (parameter->address);

			if (parameter->indirect) {

				/*----------------------*/
				/*	Varparameter	*/
				/*----------------------*/

				Consume (sident);
				Symbol argument = FindSymbol (getTokenName());

				(void) GetAddress(argument);

				VerifyType (argument->type, parameter->type);

				if (argument->storage != parameter->storage)
					Error (MsgInvalid, "segment");

				CreateStoreNode (reg, address_size (parameter->storage));

			} else {
				/*----------------------*/
				/*	Valueparameter	*/
				/*----------------------*/

				GetValue (parameter->type);
				CreateStoreNode (reg, typesize(parameter->type));
			}
			CompileTree();
		}
		if (sym->size) Consume(srkz);

		CodeCallAddr (sym->address, 0);
		return;
	}

	/*----------------------*/
	/*	Zuweisung	*/
	/*----------------------*/

	uint16  size = GetAddress(sym);
	Storage stor = sym->storage;

	if (sym->type==word && size==2 && getToken()==scaret) {
		Consume(scaret);
		CreateLoadNode (stor, size);
		stor = external;
		size = 1;
	}
	Consume (sassign);
	GetValue(sym->type);
	CreateStoreNode (stor, size);
	CompileTree();
}


/*----------------------------------------------*\
|*						*|
|*	Increment / Decrement			*|
|*						*|
\*----------------------------------------------*/

static void
DoIncr (void) {

	Consume(sincr);
	Consume(srka);
	Consume(sident);

	Symbol sym = FindSymbol(getTokenName());

	uint16 size = GetAddress (sym);

	CreateNode (mincr1u, address_size(sym->storage), size, FALSE);
	getLastInserted()->value = sym->storage;

	CompileTree();

	Consume(srkz);
}


static void
DoDecr (void) {

	Consume(sdecr);
	Consume(srka);
	Consume(sident);

	Symbol sym = FindSymbol(getTokenName());

	uint16 size = GetAddress (sym);

	CreateNode (mdecr1u, address_size(sym->storage), size, FALSE);
	getLastInserted()->value = sym->storage;

	CompileTree();

	Consume(srkz);
}


/*----------------------------------------------*\
|*						*|
|*	Shift left / shift right		*|
|*						*|
\*----------------------------------------------*/

static void
DoShl (void) {

	Consume (sshl);
	Consume (srka);
	Consume (sident);

	Symbol sym = FindSymbol(getTokenName());
	uint16 size = GetAddress (sym);

	CreateNode (mshl1u, address_size(sym->storage), size, FALSE);
	getLastInserted()->value = sym->storage;

	CompileTree();

	Consume(srkz);
}

static void
DoShr (void) {

	Consume (sshr);
	Consume (srka);
	Consume (sident);

	Symbol sym = FindSymbol(getTokenName());
	uint16 size = GetAddress (sym);

	CreateNode (mshr1u, address_size(sym->storage), size, FALSE);
	getLastInserted()->value = sym->storage;

	CompileTree();

	Consume(srkz);
}


/*----------------------------------------------*\
|*						*|
|*	Bitmanipulationen			*|
|*						*|
\*----------------------------------------------*/

static void
DoBit (unsigned cmd) {

	Cns();
	Consume(srka);
	Consume(sident);

	Symbol sym = FindSymbol(getTokenName());

	if (sym->sclass==bitvariable) {

		CreateConstantNode (sym->address & 0xff);	/* byte@	*/
		CreateConstantNode ((sym->address>>8) & 0x07);	/* bit@		*/
		if (sym->address&0x8000 && cmd<3) cmd=3-cmd;	/* complement	*/

	} else {
		GetAddress(sym);
		Consume(scomma);
		GetValue(word);
	}
	Consume(srkz);

	switch (cmd) {
	case 1:	CreateNode (mbitset, address_size(sym->storage), 1, FALSE);
		break;
	case 2:	CreateNode (mbitclr, address_size(sym->storage), 1, FALSE);
		break;
	case 3:	CreateNode (mbitcpl, address_size(sym->storage), 1, FALSE);
		break;
	}
	getLastInserted()->value = sym->storage;
	CompileTree();
}


/*----------------------------------------------*\
|*						*|
|*	Inline Bytestrom
|*						*|
\*----------------------------------------------*/

static void
DoInline (void) {

	Consume (sinline);
	Consume (srka);

	for (;;) {

		uint8 byte1 = ConstantValueOfType(word);
		uint8 byte2;

		switch (CodeSize[byte1]) {

		case 1:
			Code1 (byte1);
			break;
		case 2:
			Consume (scomma);
			Code2 (byte1, ConstantValueOfType(word));
			break;
		default:
			Consume (scomma);
			byte2 = ConstantValueOfType(word);
			Consume (scomma);
			Code3 (byte1, byte2, ConstantValueOfType(word));
		}

		if (getToken() != scomma) break;
		Consume (scomma);
	}
	Consume(srkz);
}


/*==============================================================*\
|*								*|
|*								*|
|*	Nichtelementare Befehle ohne Sprünge und Schleifen 	*|
|*								*|
|*								*|
\*==============================================================*/


/*----------------------------------------------*\
|*						*|
|*	Befehlssequenz mit gesperrtem Interrupt	*|
|*						*|
\*----------------------------------------------*/

static void
DoLock (void) {

	Consume (slock);

	Code2(0xc0, 0xa8);	/* PUSH PSW	*/
	Code2(0xc2, 0xaf);	/* CLR IEN	*/

	DoStatementSequence (FALSE, 0, 0);

	Code2(0xd0, 0xa8);	/* POP PSW	*/

	Consume(send);
}


/*==============================================================*\
|*								*|
|*								*|
|*		Bedingte Befehle (IF und CASE)			*|
|*								*|
|*								*|
\*==============================================================*/


static void
DoStatement (BOOL rok, Label*exit, Label*cont);


/*----------------------------------------------*\
|*						*|
|*	IF THEN ELSE				*|
|*						*|
\*----------------------------------------------*/

static void
DoIf (BOOL rok, Label*exit, Label*cont) {

	Consume (sif);

	Label       l_else;
	LabelInit (&l_else);

	GetValue(byte);
	CompileCondition (&l_else, FALSE);

	Consume(sthen);

	DoStatement (rok, exit, cont);

	if (getToken() == selse) {

		Consume (selse);

		Label	    l_end;
		LabelInit (&l_end);
		CodeJump  (&l_end, 0, 0);

		CodeLabel (&l_else);

		DoStatement (rok, exit, cont);

		CodeLabel (&l_end);

	} else {
		CodeLabel (&l_else);
		return;
	}
}


/*----------------------------------------------*\
|*						*|
|*	CASE IN END				*|
|*						*|
\*----------------------------------------------*/


static void
DoCaseTagSequenz (uint8* ofsptr, Label * hit, Label * skip) {

	/*------------------------------------------------------*\
	|*	Erwartet im Acc den Wert des Caseausdruck,	*|
	|*	erhöht um den Wert von *ofsptr			*|
	|*	Parst eile List vom Wertebereichen,		*|
	|*	und erzeugt den Code, hinter die nachfolgenden	*|
	|*	zu springen, den der Caseausdruck in einem der	*|
	|*	Bereiche liegt.					*|
	\*------------------------------------------------------*/

	for (;;) {
		uint16 rangebase = ConstantValueOfType(word);
		uint16 rangesize = 1;
		if (getToken() == supto) {
			Cns();
			rangesize = ConstantValueOfType(word) - rangebase + 1;
			if (rangesize<1 || rangesize > 255) {
				Error(MsgRestriction, "case:range");
				rangesize = 1;
			}
		}

		/*----------------------------------------------*\
		|*	zum Acc muß adjust addiert werden,	*|
		|*	damit bei Case-Ausdruck = rangebase	*|
		|*	Acc + Adjust == 0			*|
		\*----------------------------------------------*/

		uint8 adjust = -rangebase - *ofsptr;

		switch (adjust) {
		case 0x01:
			Code1 (0x04);		/* INC A	*/
			break;
		case 0xff:
			Code1 (0x14);		/* DEC A	*/
		case 0x00:
			break;			/* NOP		*/
		default:
			Code2 (0x24, adjust);	/* ADD A,#	*/
		}

		*ofsptr += adjust;

		BOOL lastlabel = getToken() != scomma;

		/*----------------------------------------------*\
		|*	Code zur Sprungbestimmung		*|
		\*----------------------------------------------*/

		if (rangesize>=2) {
			*ofsptr -= rangesize;
			Code2 (0x24, 0x100-rangesize);		/* ADD A,#data	*/
			if (lastlabel && getOption('j')==1) {
				CodeJump (skip, 0x40, 0);	/* JC		*/
			} else {
				CodeJump (hit, 0x50, 0);	/* JNC		*/
			}
		} else {
			if (lastlabel && getOption('j')==1) {
				CodeJump (skip, 0x70, 0);	/* JNZ		*/
			} else {
				CodeJump (hit, 0x60, 0);	/* JZ		*/
			}
		}

		/*----------------------------------------------*\
		|*	Code zur Sprungbestimmung		*|
		\*----------------------------------------------*/

		if (lastlabel) {
			if (getOption('j')!=1) CodeJump (skip, 0, 0);
			break;
		}

		Consume (scomma);
	}
}


static void
DoCase (BOOL rok, Label*exit, Label*cont) {

	Consume (scase);

	CreateConstantNode (ACC);
	GetValue (word);
	CreateStoreNode (reg, 1);
	CompileTree();

	Consume (sof);

	Label esac;
	LabelInit (&esac);

	uint8	offset = 0;

	do {
		Label skip;
		Label hit;

		LabelInit (&skip);
		LabelInit (&hit);

		DoCaseTagSequenz(&offset, &hit, &skip);

		CodeLabel (&hit);

		Consume(scolon);

		DoStatement (rok, exit, cont);
		Consume(ssemi);

		CodeJump  (&esac, 0, 0);
		CodeLabel (&skip);

	} while (getToken() != send && getToken() != selse);

	if (getToken() == selse) {
		Consume (selse);
		DoStatementSequence (rok, exit, cont);
	}

	CodeLabel  (&esac);

	Consume(send);
}


/*==============================================================*\
|*								*|
|*								*|
|*	Schleifenbefehle (WHILE, REPEAT, LOOP, FOR)		*|
|*								*|
|*								*|
\*==============================================================*/


/*----------------------------------------------*\
|*						*|
|*	WHILE DO				*|
|*						*|
\*----------------------------------------------*/

static void
DoWhile (BOOL rok) {

	Consume (swhile);

	Label cont;
	Label exit;

	LabelInit (&cont);
	LabelInit (&exit);

	CodeLabel (&cont);

	GetValue(byte);
	CompileCondition (&exit, FALSE);

	Consume(sdo);

	DoStatement (rok, &exit, &cont);

	CodeJump (&cont, 0, 0);
	CodeLabel  (&exit);
}


/*----------------------------------------------*\
|*						*|
|*	REPEAT UNTIL				*|
|*						*|
\*----------------------------------------------*/

static void
DoRepeat (BOOL rok) {

	Consume (srepeat);

	Label cont;
	Label exit;

	LabelInit (&cont);
	LabelInit (&exit);

	CodeLabel  (&cont);

	DoStatementSequence (rok, &exit, &cont);

	Consume(suntil);

	GetValue(byte);
	CompileCondition (&cont, FALSE);

	CodeLabel  (&exit);
}


/*----------------------------------------------*\
|*						*|
|*	LOOP END (EXIT)				*|
|*						*|
\*----------------------------------------------*/


static void
DoLoop (BOOL rok) {

	Consume (sloop);

	Label cont;
	Label exit;

	LabelInit (&cont);
	LabelInit (&exit);

	CodeLabel  (&cont);

	DoStatementSequence (rok, &exit, &cont);

	CodeJump (&cont, 0, 0);
	CodeLabel  (&exit);

	Consume(send);
}


/*----------------------------------------------*\
|*						*|
|*	FOR := TO DO				*|
|*						*|
\*----------------------------------------------*/

static void
DoFor (BOOL rok) {

	Consume (sfor);

	Label loop;
	Label cont;
	Label exit;
	LabelInit (&loop);
	LabelInit (&cont);
	LabelInit (&exit);

	/*---------------------------------------------*\
	|	Laufvariable				|
	\*---------------------------------------------*/

	Consume (sident);

	Symbol loopvar = FindSymbol(getTokenName());
	if (loopvar->sclass != variable || loopvar->indirect)
		Error (MsgExpectedVariable, 0);
	if (loopvar->type != word) VerifyType (loopvar->type, byte);
	unsigned loopvarsize = typesize (loopvar->type);

	/*---------------------------------------------*\
	|	Initialisiere mit Startwert		|
	\*---------------------------------------------*/

	CreateConstantNode (loopvar->address);

	Consume(sassign);
	Node start = GetValue(word);
	BOOL   start_const = start->op == cval;
	uint16 start_value = start->value;
	CreateStoreNode (loopvar->storage, loopvarsize);
	CompileTree();

	/*---------------------------------------------*\
	|	Zaehlvariable				|
	\*---------------------------------------------*/

	static unsigned cnum;
	char id[12];
	snprintf (id, 12, "<for:%u>", ++cnum);
	Symbol counter = CreateSymbol (variable, id);
	counter->storage = internal;
	counter->type    = byte;
	AllocateMemory (counter);
	unsigned countvarsize = typesize (counter->type);

	CreateConstantNode (counter->address);

	/*---------------------------------------------*\
	|	HV := ende - (start-1) ;start=const	|
	|	HV := ende+1 - (start) ;start=var	|
	\*---------------------------------------------*/

	Consume(sto);
	GetValue(word);

	if (start_const) {
		CreateConstantNode (start_value - 1);
	} else {
		CreateConstantNode (1);
		CreateNode (addu, loopvarsize, 0, FALSE);
		CreateConstantNode (loopvar->address);
		CreateLoadNode (loopvar->storage, loopvarsize);
	}
	CreateNode (subu, countvarsize, 0, FALSE);
	CreateStoreNode (counter->storage, countvarsize);
	CompileTree();

	/*---------------------------------------------*\
	|	Schleifenkommenado			|
	\*---------------------------------------------*/

	Consume(sdo);

	CodeLabel (&loop);

	DoStatement (rok, &exit, &cont);

	CodeLabel (&cont);

	/*---------------------------------------------*\
	|	Erhöhe die Laufvariable			|
	\*---------------------------------------------*/

	CreateConstantNode (loopvar->address);
	CreateNode (mincr1u, loopvarsize, 0, FALSE);
	getLastInserted()->value=loopvar->storage;
	CompileTree();

	/*---------------------------------------------*\
	|	Erniedige die Hilfsvariable		|
	|	und springe wenn 0 erreicht		|
	\*---------------------------------------------*/

	uint16 size = GetAddress (counter);

	CreateNode (mdecr1u, address_size(counter->storage), size, FALSE);
	getLastInserted()->value = counter->storage;

	CompileCondNonZero (&loop);

	CodeLabel (&exit);
}


/*==============================================================*\
|*								*|
|*								*|
|*	Statement und StatementSequence				*|
|*								*|
|*								*|
\*==============================================================*/

static void
DoStatement (BOOL rok, Label*exit, Label* cont) {

	switch (getToken()) {

	/*------------------------------*\
	|*	Elementare Befehle	*|
	\*------------------------------*/

	case sident:
		DoAssignOrCall();
		return;

	case sincr:
		DoIncr();
		return;

	case sdecr:
		DoDecr();
		return;

	case sshl:
		DoShl ();
		return;

	case sshr:
		DoShr ();
		return;

	case sset:
		DoBit(1);
		return;

	case sclr:
		DoBit(2);
		return;

	case scpl:
		DoBit(3);
		return;

	case sinline:
		DoInline();
		return;

	/*------------------------------*\
	|*	Ohne Spruenge		*|
	\*------------------------------*/

	case slock:
		DoLock();
		return;

	/*------------------------------*\
	|*	Bedingte Befehle	*|
	\*------------------------------*/

	case sif:
		DoIf(rok, exit, cont);
		return;

	case scase:
		DoCase(rok, exit, cont);
		return;

	/*------------------------------*\
	|*	Schleifen		*|
	\*------------------------------*/

	case swhile:
		DoWhile (rok);
		return;

	case srepeat:
		DoRepeat(rok);
		return;

	case sloop:
		DoLoop(rok);
		return;

	case sfor:
		DoFor(rok);
		return;

	/*------------------------------*\
	|*	Spruenge		*|
	\*------------------------------*/

	case sbreak:
	case sexit:
		if (!exit) break;
		Cns();
		CodeJump (exit, 0, 0);
		return;

	case scontinue:
		if (!cont) break;
		Consume(scontinue);
		CodeJump (cont, 0, 0);
		return;

	/*------------------------------*\
	|*	Return			*|
	\*------------------------------*/

	case sreturn:
		if (!rok) break;
		Consume(sreturn);
		DoLeave();
		return;

	/*------------------------------*\
	|*	Leerer Befehl		*|
	\*------------------------------*/

	case ssemi:
	case send:
	case selse:
	case suntil:
		return;

	/*------------------------------*\
	|*	Befehlsfolge		*|
	\*------------------------------*/

	case sbegin:
		Consume(sbegin);
		DoStatementSequence (rok, exit, cont);
		Consume(send);
		return;

	default:
		break;
	}

	Fatal (MsgExpectedStatement, 0);
}


void
DoStatementSequence (BOOL rok, Label*exit, Label*cont) {

	for (;;) {
		DoStatement (rok, exit, cont);
		if (getToken() != ssemi) break;
		Consume(ssemi);
	}
}


/*===================================================================*\
|*	$Id: statements.c,v 1.1 2018/12/21 00:15:19 wolf Exp $
\*===================================================================*/
