/****************************************************************)
(*								*)
(*								*)
(*	N       N        AAA       N       N      OOOOO		*)
(*	NN      N       A   A      NN      N     O     O	*)
(*	N N     N      A     A     N N     N    O       O	*)
(*	N  N    N     A       A    N  N    N    O       O	*)
(*	N   N   N     A       A    N   N   N    O       O	*)
(*	N    N  N     AAAAAAAAA    N    N  N    O       O	*)
(*	N     N N     A       A    N     N N    O       O	*)
(*	N      NN     A       A    N      NN     O     O	*)
(*	N       N     A       A    N       N      OOOOO		*)
(*								*)
(*								*)
(****************************************************************)
(*	Compiler fuer ein *sehr* eingeschraenktes Pascal.	*)
(*	Erzeugt Maschinencode fuer 8031 & co Prozessoren.	*)
(****************************************************************/

/*===================================================================*\
|*	$Id: symboltable.c,v 1.1 2018/12/21 00:15:19 wolf Exp $
|*===================================================================*|
|*
|*	$Log: symboltable.c,v $
|*	Revision 1.1  2018/12/21 00:15:19  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 22:27:22  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 16:03:56  wolf
|*	Initial revision
|*
|*	Revision 1.4  2011/08/09 17:26:17  wolf
|*	* Cleanup
|*
|*	Revision 1.3  2010/12/24 16:26:09  wolf
|*	*** empty log message ***
|*
|*	Revision 1.2  2008/12/12 20:22:21  wolf
|*	*** empty log message ***
|*
|*	Revision 1.1  2008/12/07 09:43:42  wolf
|*	Initial revision
|*
|*===================================================================*|
|*	1988-1996 als Turbopascal-Programm für 8080 und 8086
\*===================================================================*/


#include "nano.h"

#include "symboltable.h"

#include <string.h>
#include <strings.h>
#include <stdlib.h>


/*==============================================================*\
|*								*|
|*		Symboltabelle und Zeiger in diese Tabelle	*|
|*								*|
\*==============================================================*/

static
SymbolTableEntry Symbols[MaxSymbols];

static
SymbolTableEntry* EndOfTable = Symbols;

static
SymbolTableEntry* Released = NULL;


/*==============================================================*\
|*								*|
|*	Trage neues Symbol in die Symboltabelle ein		*|
|*								*|
\*==============================================================*/

Symbol
CreateSymbol (Class newclass, const char * newid) {

	/*--------------------------------------------------------------*\
	|*	Schaue nach, ob das Symbol bereits definiert ist	*|
	\*--------------------------------------------------------------*/

	SymbolTableEntry* p = EndOfTable;
	while (p>Symbols) {
		p--;
		if (p->sclass==noclass) break;
		if (p->sclass>noclass && !strcasecmp(p->id, newid)) {
			Fatal (MsgNameDefined, newid);
			return NULL;
		}
	}

	/*----------------------------------------------*\
	|*	Noch Platz in der Tabelle?		*|
	\*----------------------------------------------*/

	if (EndOfTable >= &Symbols[MaxSymbols]) {
		Error (MsgLimit, "MaxSymbols");
		return NULL;
	}

	/*----------------------------------------------*\
	|*	Kopiere den Identifier			*|
	\*----------------------------------------------*/

	char * id = (char*) malloc (strlen(newid)+1);
	strcpy (id, newid);

	/*----------------------------------------------*\
	|*	Initialisiere das Symbol?		*|
	\*----------------------------------------------*/

	p = EndOfTable++;
	p->id		= id;
	p->sclass	= newclass;
	p->type		= notype;
	p->storage	= internal;
	p->absolute	= FALSE;
	p->used		= FALSE;
	p->indirect	= FALSE;
	p->address	= 0;
	p->size		= 0;
	return p;
}


/*==============================================================*\
|*								*|
|*	Suche einen Identifier in der Symboltabelle		*|
|*								*|
\*==============================================================*/

Symbol
FindSymbol (const char oldid[]) {

	SymbolTableEntry* p = EndOfTable;
	while (p>Symbols) {
		p--;
		if (p->sclass>noclass && !strcasecmp(p->id, oldid)) {
			p->used = TRUE;
			return p;
		}
	}
	Error (MsgNameUndefined, oldid);
	return NULL;
}


/*----------------------------------------------*\
|*	Symbol bekannt (als was auch immer)?	*|
\*----------------------------------------------*/

BOOL
isDefined (const char oldid[]) {
	if (!oldid) return FALSE;
	Symbol p;
	for (p=Symbols; p<EndOfTable; p++)
		if (p->sclass>noclass && !strcasecmp(p->id, oldid))
			return TRUE;
	return FALSE;
}


/*==============================================================*\
|*								*|
|*	Suche eine Variable rückwärts über ihre Adresse		*|
|*								*|
\*==============================================================*/

Symbol
getSymbolByAddress (uint16 addr) {

	SymbolTableEntry* p = EndOfTable;

	while (p>Symbols) {
		p--;
		if (p->storage != internal)
			continue;
		if (p->sclass != variable && p->sclass != parameter)
			continue;
		if (p->address==addr)
			return p;
		if (p->address+1==addr && p->type!=byte && !p->indirect)
			return p;
	}
	return NULL;
}


/*==============================================================*\
|*								*|
|*	Zeige einen Eintrag der Symboltabelle an		*|
|*								*|
\*==============================================================*/

void
ShowSymbol (Symbol sym, FILE * file, const char prefix[]) {

	unsigned address = sym -> address;

	char		used	= sym->used ? ' ' : '-';
	char		ext1	= ' ';
	char		ext2	= sym->indirect ? '^' : ' ';
	char		abs	= sym->absolute ? '=' : ' ';
	char		asize[7] = {0};
	const char *	cname	= "     ";
	const char *	tname	= "      ";
	const char *	sname	= "     ";

	if (sym->size) snprintf (asize, 7, "[%u]", sym->size);

	switch (sym->type) {
	case bit:
		tname = "bit+  ";
		break;
	case byte:
		tname = "byte  ";
		break;
	case word:
		tname = "word  ";
		break;
	case string:
		tname = "string";
		break;
	case proc:
		tname = "proc  ";
		break;
	default:;
	}

	switch (sym->sclass) {
	case constant:
		cname="const";
		break;
	case parameter:
		cname="param";
		break;
	case bitvariable:
		ext1 = '.';
		ext2 = ((address >> 8) & 0x07) + '0';
		if (address&0x8000) tname = "bit-  ";
		address &= 0xff;
	case variable:
		cname="var  ";
		break;
	default:;
	}

	switch (sym->storage) {
	case external:
		sname = "xdata";
		break;
	case code:
		sname = "code ";
		break;
	case internal:
		sname = "data ";
		break;
	case reg:
		sname = "reg  ";
		break;
	default:;
	}

	fprintf (file, "%s%c%s %s %s %c%04x%c%c %s%s\n",
		prefix,
		used,
		cname, tname, sname,
		abs, address, ext1, ext2,
		sym->id, asize);
}

/*------------------------------------------------------*\
|*	Zeige einen gerade freigegebenen Scope an	*|
\*------------------------------------------------------*/

void
ShowScope (Symbol scope, FILE * file) {
	fprintf (file, "\t======================================================\n");
	ShowSymbol (scope, file, "\t");
	if (scope+1<Released)
	fprintf (file, "\t------------------------------------------------------\n");
	Symbol p;
	for (p=scope+1; p<Released; p++)
		ShowSymbol (p, file, "\t");
	fprintf (file, "\t======================================================\n");
}


/*==============================================================*\
|*								*|
|*	Gebe das Ende der Symboltabelle frei			*|
|*								*|
\*==============================================================*/

void
ReleaseScope(Symbol sym, unsigned npara) {
	sym->sclass = constant;
	for (unsigned i=1; i<=npara; i++)
		sym[i].sclass = parameter;
	Released	= EndOfTable;
	EndOfTable	= sym+1+npara;

	if (chkOption('l','s')) {
		lstfileselect ("symbols");
		ShowScope (sym, lstfile);
	}

	for (SymbolTableEntry* p=EndOfTable; p<Released; p++) {

		free((void*) p->id);
		p->id = NULL;
	}
}


/*===================================================================*\
|*	$Id: symboltable.c,v 1.1 2018/12/21 00:15:19 wolf Exp $
\*===================================================================*/
