/****************************************************************)
(*								*)
(*								*)
(*	N       N        AAA       N       N      OOOOO		*)
(*	NN      N       A   A      NN      N     O     O	*)
(*	N N     N      A     A     N N     N    O       O	*)
(*	N  N    N     A       A    N  N    N    O       O	*)
(*	N   N   N     A       A    N   N   N    O       O	*)
(*	N    N  N     AAAAAAAAA    N    N  N    O       O	*)
(*	N     N N     A       A    N     N N    O       O	*)
(*	N      NN     A       A    N      NN     O     O	*)
(*	N       N     A       A    N       N      OOOOO		*)
(*								*)
(*								*)
(****************************************************************)
(*	Compiler fuer ein *sehr* eingeschraenktes Pascal.	*)
(*	Erzeugt Maschinencode fuer 8031 & co Prozessoren.	*)
(****************************************************************/

/*===================================================================*\
|*	$Id: tree.c,v 1.1 2018/12/21 00:15:19 wolf Exp $
|*===================================================================*|
|*
|*	$Log: tree.c,v $
|*	Revision 1.1  2018/12/21 00:15:19  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 22:27:22  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 16:03:56  wolf
|*	Initial revision
|*
|*	Revision 1.4  2011/08/09 17:26:17  wolf
|*	* Cleanup
|*
|*	Revision 1.3  2008/12/12 22:30:58  wolf
|*	*** empty log message ***
|*
|*	Revision 1.2  2008/12/12 20:22:21  wolf
|*	*** empty log message ***
|*
|*===================================================================*|
|*	1988-1996 als Turbopascal-Programm für 8080 und 8086
\*===================================================================*/


#include "nano.h"
#include "tree.h"

#include <string.h>
#include <stdio.h>


const char * OperatorName[] = { OPNAMES };

/*==============================================================*\
|*								*|
|*	Tabelle der Ausdrucks-Baum-Knoten			*|
|*								*|
\*==============================================================*/

static
NodeTableEntry	Nodes[MaxNodes];


/*==============================================================*\
|*								*|
|*	Hilfszeiger						*|
|*								*|
\*==============================================================*/

static
Node	EndOfNodeTable=Nodes;

static
Node	ReferenceCount=Nodes;

static
Node	LastInserted=NULL;

static
Node	MaximumUsed=Nodes;


/*==============================================================*\
|*								*|
|*	Liefere Zeiger auf den zuletzt eingefügten Knoten	*|
|*								*|
\*==============================================================*/

Node
getLastInserted (void) {
	return LastInserted;
}


/*==============================================================*\
|*								*|
|*	Liefere Zeiger auf den jüngsten unbenutzten Knoten,	*|
|*	und markiere ihne als benutzt				*|
|*								*|
\*==============================================================*/

Node
getLastUnused (void) {

	Node p = EndOfNodeTable;

	while (p>Nodes) {
		p--;
		if (p->used) continue;
		p->used = TRUE;
		++ReferenceCount;
		return p;
	}

	Fatal (MsgInternal, "getLastUnused");
	return NULL;
}


/*==============================================*\
|*						*|
|*	Dumproutinen				*|
|*						*|
\*==============================================*/

void
DumpNode (Node node, FILE * file) {
	fprintf (file, "#%02li:\t%s in=%u out=%u",
		node-Nodes,
		OperatorName[node->op],
		node->isize,
		node->osize);
	if (node->used) fprintf (file, " USED");
	if (node->op==cval) fprintf (file, " value=%02x",
		node->value);
	if (node->left) fprintf (file, " left=%li", node->left-Nodes);
	if (node->right) fprintf (file, " right=%li", node->right-Nodes);
	if (node->comment) fprintf (file, " *** %s ***",
		node->comment);
	fprintf (file, "\n");
}

void
DumpNodes (FILE * file) {
	for (Node node=Nodes; node<EndOfNodeTable; node++)
		DumpNode (node, file);
}

void
TreeDump (Node node, FILE * file, char c, unsigned indent) {
	fprintf (file, "%*s%c: %s",
		indent*4, "", c, OperatorName[node->op]);
	if (node->isize) fprintf (file, " in=%u", node->isize);
	if (node->osize) fprintf (file, " out=%u", node->osize);
	if (node->op==cval) fprintf (file, " value=%x", node->value);
	fprintf (file, " #reg=%d", node->nregs);
	if (node->sregs) fprintf (file, " sfr=%x", node->sregs);
	fprintf (file, " #%li", node-Nodes);
	if (node->comment) fprintf (file, " *** %s ***", node->comment);
	fprintf (file, "\n");
	if (node->left ) TreeDump (node->left , file, 'L', indent+1);
	if (node->right) TreeDump (node->right, file, 'R', indent+1);
}


/*==============================================================*\
|*								*|
|*	Erzeuge neuen Knoten (bzw. alloziere in Tabelle)	*|
|*								*|
|*	Als Kinder werden automatisch die jüngsten unbenutzen	*|
|*	Knoten in der Tabelle eingehängt			*|
|*								*|
\*==============================================================*/


void
CreateNode (Operator op, unsigned isize, unsigned osize, BOOL flip) {

	/*----------------------------------------------*\
	|*	Autoreset der Tabelle			*|
	\*----------------------------------------------*/

	if (ReferenceCount>=EndOfNodeTable) {
		EndOfNodeTable = Nodes;
		ReferenceCount = Nodes;
	}

	/*----------------------------------------------*\
	|*	Reserviere einen Eintrag in der Tabelle	*|
	\*----------------------------------------------*/

	if (EndOfNodeTable >= &Nodes[MaxNodes]) {
		Error (MsgLimit, "MaxNodes");
		return;
	}

	LastInserted = EndOfNodeTable;

	/*----------------------------------------------*\
	|*	Initialisiere den Eintrag		*|
	\*----------------------------------------------*/

	LastInserted->op	= op;
	LastInserted->osize	= osize;
	LastInserted->isize	= isize;
	LastInserted->value	= 0;
	LastInserted->used	= FALSE;
	LastInserted->nregs	= -99;
	LastInserted->sregs	= 0;
	LastInserted->left	= NULL;
	LastInserted->right	= NULL;
	LastInserted->comment	= NULL;

	/*----------------------------------------------*\
	|*	Trage die Kinder ein			*|
	\*----------------------------------------------*/

	if (op>_1r_)
		LastInserted->right = getLastUnused();
	if ((op>_1l_ && op<_1r_) || op>_2_)
		LastInserted->left = getLastUnused();

	if (flip) {
		Node n = LastInserted->left;
		LastInserted->left = LastInserted->right;
		LastInserted->right = n;
	}

	/*----------------------------------------------*\
	|*	Berechne die Insize			*|
	\*----------------------------------------------*/

	if (isize == 0) {
		if (LastInserted->right && LastInserted->right->osize>LastInserted->isize)
			LastInserted->isize = LastInserted->right->osize;
		if (LastInserted->left && LastInserted->left->osize>LastInserted->isize)
			LastInserted->isize = LastInserted->left->osize;
	}

	if (isize == 0xff) {
		if (LastInserted->right)
			LastInserted->isize = LastInserted->right->osize;
		if (LastInserted->left && LastInserted->left->osize<LastInserted->isize)
			LastInserted->isize = LastInserted->left->osize;
	}

	if (!LastInserted->osize) LastInserted->osize=LastInserted->isize;

	/*----------------------------------------------*\
	|*	Markiere als belegt.			*|
	\*----------------------------------------------*/

	if (++EndOfNodeTable>MaximumUsed) MaximumUsed = EndOfNodeTable;

}


/*==============================================================*\
|*								*|
|*	Erzeuge neuen Knoten mit op=cval			*|
|*								*|
|*	Trage den uebergebenen Wert in das Feld "value" ein.	*|
|*								*|
\*==============================================================*/

void
CreateConstantNode (unsigned value) {
	CreateNode (cval, 0, value>=256 ? 2 : value ? 1 : 0, FALSE);
	LastInserted->value=value;
}


/*==============================================================*\
|*								*|
|*	Erzeuge einen Abspeicherknoten				*|
|*								*|
\*==============================================================*/

void
CreateStoreNode (Storage storage, unsigned osize) {
	CreateNode (mwrite, address_size(storage), osize, FALSE);
	LastInserted->value = storage;
}


/*==============================================================*\
|*								*|
|*	Erzeuge einen Ausleseknoten				*|
|*								*|
\*==============================================================*/

void
CreateLoadNode (Storage storage, unsigned osize) {
	CreateNode (mread, address_size(storage), osize, FALSE);
	LastInserted->value = storage;
}


/*==============================================*\
|*						*|
|*	Auswertung konstanter Ausdruecke	*|
|*						*|
\*==============================================*/

static inline void
SwapChilds (Node node) {
	Node n = node->left;
	node->left = node->right;
	node->right = n;
	node->comment = "Swapped!!!";
}

static Node
FoldConstant (Node node) {

	if (!node) return node;

	BOOL lc	= node->left  && node->left ->op==cval;
	BOOL rc	= node->right && node->right->op==cval;

	if (!lc && !rc) return node;

	uint16 lv	= lc ? node->left -> value : 0;
	uint16 rv	= rc ? node->right-> value : 0;

	uint32 value = 0;

	switch (node->op) {

	/*----------------------------------------------*\
	|*	Arithmetik (zweistellige)		*|
	\*----------------------------------------------*/

	case addu:
		if (!rc) return node;
		if (lc) {
			value = (uint16) (lv + rv);
			break;
		}
		if (rv == 0) return node->left;
		return node;

	case subu:
		if (!rc) return node;
		if (lc) {
			value = (uint16) (lv - rv);
			break;
		}
		if (rv == 0) return node->left;
		return node;

	/*----------------------------------------------*\
	|*	Mul/Div/Mod 				*|
	|*	Beschraenkung der Groessen (8031)	*|
	\*----------------------------------------------*/

	case mulu:
		if (rc) {
			if (lc) {
				value = lv * rv;
				break;
			}
			if (rv == 0) return node->right;
			if (rv == 1) return node->left;
		}
		node->isize = 1;
		return node;

	case divu:
		if (lc && lv == 0) return node->left;
		if (rc) {
			if (lc) {
				value = rv ? lv / rv : 0xffffffff;
				break;
			}
			if (rv == 1) return node->left;
		}
		node->osize = 1;
		node->isize = 1;
		return node;

	case modu:
		if (lc && lv == 0) return node->left;
		if (rc) {
			if (lc) {
				value = rv ? lv % rv : 0xffffffff;
				break;
			}
			if (rv == 1) {value=0; break; }
		}
		node->osize = 1;
		node->isize = 1;
		return node;

	/*----------------------------------------------*\
	|*	Logik (zweistellig)			*|
	\*----------------------------------------------*/

	case andu:
		if (!rc) return node;
		if (lc) {
			value = lv & rv;
			break;
		}
		if (rv == 0) return node->right;
		if (rv == 0xffff) return node->left;
		if (rv == 0xff && node->left->osize<=1) return node->left;
		return node;

	case oru:
		if (!rc) return node;
		if (lc) {
			value = lv | rv;
			break;
		}
		if (rv == 0) return node->left;
		if (rv == 0xffff) return node->right;
		if (rv == 0xff && node->left->osize<=1) return node->right;
		return node;

	case xoru:
		if (!rc) return node;
		if (lc) {
			value = lv ^ rv;
			break;
		}
		if (rv == 0) return node->left;
		return node;

	/*----------------------------------------------*\
	|*	Vergleiche				*|
	\*----------------------------------------------*/

	case eq:
		if (rc && lc) {
			value = lv == rv ? CodeTrue : 0;
			break;
		}
		return node;

	case ne:
		if (rc && lc) {
			value = lv != rv ? CodeTrue : 0;
			break;
		}
		return node;

	case ltu:
		if (rc && lc) {
			value = lv < rv ? CodeTrue : 0;
			break;
		}
		if ((lc && (lv==0xffff || (lv==0xff && node->isize<=1))) || (rc && rc==0)) {
			Warning (MsgConstCondition, "false");
			break;
		}
		if (lc && (lv == 0)) {
			node->op = ne;
			SwapChilds (node);
		}
		return node;

	case geu:
		if (rc && lc) {
			value = lv >= rv ? CodeTrue : 0;
			break;
		}
		if ((lc && (lv==0xffff || (lv==0xff && node->isize<=1))) || (rc && rc==0)) {
			Warning (MsgConstCondition, "true");
			value = CodeTrue;
			break;
		}
		if (lc && lv==0) {
			node->op = eq;
			SwapChilds (node);
		}
		return node;

	/*----------------------------------------------*\
	|*	Binaee Operationen			*|
	\*----------------------------------------------*/

	case negu:
		if (!rc) return node;
		value = ~rv;
		break;

	case notu:
		if (!rc) return node;
		value = rv ? 0 : CodeTrue;
		break;

	case shl1u:
		if (!rc) return node;
		value = rv << 1;
		break;

	default:
		return node;
	}

	if (value > 0xffff) Warning (MsgConstOverflow, 0);

	if (node->osize<=1) value &= 0xff;
	node -> op = cval;
	node -> value = value;
	node -> isize  = 0;
	node -> osize  = value>=256 ? 2 : value ? 1 : 0;
	node -> left = NULL;
	node -> right= NULL;
	return node;
}


/*======================================================*\
|*							*|
|*	Groessenreduktion der Knoten			*|
|*							*|
|*------------------------------------------------------*|
|*							*|
|*	Bei 8031-Prozessoren muessen 16-bit-Operationen	*|
|*	aus mehreren 8-bit-Operationen zusammengesetzt	*|
|*	werden. Das ist Zeit- und Speicher-aufwendig.	*|
|*	Deshalb wird 16-bit zu 8-bit reduziert,		*|
|*	wenn die oberen 8bit nicht gebraucht werden.	*|
|*							*|
\*======================================================*/

static inline uint8
imax (int a, int b) { return a>b ? a : b; }

void
OptimizeNodes (void) {

	/*==============================================*\
	|*	Phase 1 top down Größenreduktion	*|
	\*==============================================*/

	for (Node node=EndOfNodeTable-1; node>=Nodes; node--) {

		unsigned lsize = node->isize;
		unsigned rsize = node->isize;

		switch (node->op) {

		/*----------------------------------------------*\
		|*	Leseknoten (r=adresse)			*|
		\*----------------------------------------------*/
		case mread:
			break;

		/*----------------------------------------------*\
		|*	Speicherknoten (l=adresse r=value)	*|
		\*----------------------------------------------*/
		case mwrite:
		case mbitset:
		case mbitclr:
		case mbitcpl:
		case mshl1u:
		case mshr1u:
		case mincr1u:
		case mdecr1u:
			rsize = node->osize;
			break;

		/*----------------------------------------------*\
		|*	Operatoren mit isize !<= osize		*|
		\*----------------------------------------------*/
		case addu:
		case subu:
		case adduc:
		case subub:
		case andu:
		case oru:
		case xoru:
		case mulu:
		case cplu:
		case negu:
		case bitis0:
		case bitis1:
		case shl1u:
			if (node->isize <= node->osize) break;
			node->isize = lsize = rsize = node->osize;
			node->comment = "iSize reduced";
			break;

		/*----------------------------------------------*\
		|*	Operatoren mit isize (unabh.) osize	*|
		\*----------------------------------------------*/
		case divu:
		case modu:
			break;

		/*----------------------------------------------*\
		|*	Vergleiche: isize (unabh.) osize	*|
		\*----------------------------------------------*/
		case eq:
		case ne:
		case ltu:
		case geu:
		case notu:
			break;

		/*----------------------------------------------*\
		|*	Konstanten				*|
		\*----------------------------------------------*/
		case cval:
			if (node->osize<=1) node->value &= 0xff;
			break;

		/*----------------------------------------------*\
		|*	Blattknoten				*|
		\*----------------------------------------------*/
		case _0_:
		case _1l_:
		case _1r_:
		case _2_:
		case svoid:
			break;
		}


		/*==============================================*\
		|*	Passe osize der Kinder an eig. isize an	*|
		\*==============================================*/

		if (node->right && node->right->osize>rsize) {
			node->right->osize=rsize;
			node->comment = "oSize reduced";
		}
	
		if (node->left && node->left->osize>lsize) {
			node->left->osize=rsize;
			node->comment = "oSize reduced";
		}
	}

	/*==============================================*\
	|*	Phase 2 bottom up:			*|
	|*	- konstante Auswertung			*|
	\*==============================================*/

	for (Node node=Nodes; node<EndOfNodeTable; node++) {
		node->left = FoldConstant (node->left );
		node->right= FoldConstant (node->right);

		/*----------------------------------------------*\
		|*	In diesem Knoten benutzte Register	*|
		\*----------------------------------------------*/

		uint8	sregs = sreg_acc;
		int	nregs = 0;

		switch (node->op) {

		/*----------------------------------------------*\
		|*	Speicher				*|
		\*----------------------------------------------*/

		case mread:
			switch ((Storage) node->value) {

			case external:
			case code:
				sregs = sreg_dptr| sreg_acc;
				nregs = imax (node->left->nregs, 0);
				break;
			case internal:
			case reg:
				sregs = node->left->nregs<0 ? 0 : sreg_ri;
				nregs = imax (node->left->nregs, -1);
				break;
			default: ;
			}
			break;

		/*----------------------------------------------*\
		|*	Symmetrische Operationen		*|
		\*----------------------------------------------*/
		case addu:
		case adduc:
		case andu:
		case oru:
		case xoru:
		case mulu:
		case eq:
		case ne:
			if ((node->left->nregs < node->right->nregs &&
				node->left->nregs<0)
			||
				(node->left->nregs > node->right->nregs &&
					node->right->nregs>=0))
				SwapChilds (node);
			/* fall thru */

		/*----------------------------------------------*\
		|*	Unsymmetrische Operationen		*|
		\*----------------------------------------------*/
		case subu:
		case subub:
		case divu:
		case modu:
		case bitis0:
		case bitis1:
		case ltu:
		case geu:
			if (node->right->nregs>= 0) {
				nregs = imax ( node->right->nregs,
				node->right->osize+node->left->nregs);
			} else {
				nregs = node->left->nregs;
			}
			if(node->left->op!=cval || node->right->op!=cval) {
				switch (node->op) {
				case divu:
				case modu:
					node->osize=1;
					/* fall thru */
				case mulu:
					node->isize=1;
				default: ;
				}
			}
			break;

		/*----------------------------------------------*\
		|*	Einstellige Operationen			*|
		\*----------------------------------------------*/
		case cplu:
		case negu:
		case shl1u:
		case notu:
			nregs = imax (node->right->nregs, 0);
			break;

		/*----------------------------------------------*\
		|*	Konstanter Knoten			*|
		\*----------------------------------------------*/
		case cval:
			sregs = 0;
			nregs = -2;
			break;

		/*----------------------------------------------*\
		|*	Wurzel und Blattknoten			*|
		\*----------------------------------------------*/
		case mwrite:
		case svoid:
		case mbitset:
		case mbitclr:
		case mbitcpl:
		case mdecr1u:
		case mincr1u:
		case mshl1u:
		case mshr1u:
		case _0_:
		case _1l_:
		case _1r_:
		case _2_:
			sregs = 0;
			nregs = 0;
			break;
		}

		if (sregs && nregs<0) nregs=0;
		if (node->left ) sregs |= node->left ->sregs;
		if (node->right) sregs |= node->right->sregs;

		node-> nregs = nregs;
		node-> sregs = sregs;
	}
}


/*===================================================================*\
|*	$Id: tree.c,v 1.1 2018/12/21 00:15:19 wolf Exp $
\*===================================================================*/
