/****************************************************************)
(*								*)
(*								*)
(*	N       N        AAA       N       N      OOOOO		*)
(*	NN      N       A   A      NN      N     O     O	*)
(*	N N     N      A     A     N N     N    O       O	*)
(*	N  N    N     A       A    N  N    N    O       O	*)
(*	N   N   N     A       A    N   N   N    O       O	*)
(*	N    N  N     AAAAAAAAA    N    N  N    O       O	*)
(*	N     N N     A       A    N     N N    O       O	*)
(*	N      NN     A       A    N      NN     O     O	*)
(*	N       N     A       A    N       N      OOOOO		*)
(*								*)
(*								*)
(****************************************************************)
(*	Compiler fuer ein *sehr* eingeschraenktes Pascal.	*)
(*	Erzeugt Maschinencode fuer 8031 & co Prozessoren.	*)
(****************************************************************/

/*===================================================================*\
|*	$Id: tree.h,v 1.1 2018/12/21 00:15:19 wolf Exp $
|*===================================================================*|
|*
|*	$Log: tree.h,v $
|*	Revision 1.1  2018/12/21 00:15:19  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 22:27:54  wolf
|*	Initial revision
|*
|*	Revision 1.1  2018/12/20 16:03:56  wolf
|*	Initial revision
|*
|*	Revision 1.15  2011/08/09 17:26:58  wolf
|*	* Cleanup
|*
|*	Revision 1.14  2008/12/12 20:22:21  wolf
|*	*** empty log message ***
|*
|*	Revision 1.13  2008/12/07 09:43:42  wolf
|*	*** empty log message ***
|*
|*===================================================================*|
|*	1988-1996 als Turbopascal-Programm für 8080 und 8086
\*===================================================================*/

#ifndef __TREE_H__
#define __TREE_H__

#include "symboltable.h"
#include <stdio.h>


typedef enum {
	_0_,
	cval,
	_1l_,
	mread, mshl1u, mshr1u, mincr1u, mdecr1u,
	_1r_,
	svoid, notu, cplu, negu, shl1u,
	_2_,
	mwrite, mbitset, mbitclr, mbitcpl,
	addu, subu, adduc, subub,
	andu, oru, xoru,
	mulu, divu, modu,
	bitis0, bitis1,
	eq, ne, ltu, geu
} Operator;


#define OPNAMES \
	"0",						\
	"<const>",					\
	"1l",						\
	"mread", "m<<=1", "m>>=1", "m+=1", "m-=1",	\
	"1r",						\
	"(void)", "!", "~", "-()", "<<1",		\
	"2",						\
	"mwrite", "mbitset", "mbitclr", "mbitcpl",	\
	"+", "-", "!+", "!-",				\
	"&", "|", "^",					\
	"*", "DIV", "MOD",				\
	"bitis0", "bitis1"				\
	"==", "!=", "<", ">="

extern const char * OperatorName[];

struct node_table_entry;

typedef struct node_table_entry* Node;

typedef struct node_table_entry {
	Operator	op;
	uint8		osize;
	uint8		isize;
	uint16		value;
	BOOL		used;
	signed char	nregs;
	uint8		sregs;
	Node		left;
	Node		right;
	const char*	comment;
} NodeTableEntry;

static const uint8	sreg_acc	= 0x01;
static const uint8	sreg_dptr	= 0x20;
static const uint8	sreg_ri		= 0x10;

static const uint8 CodeTrue = 0xff;

void
CreateNode (Operator op, unsigned isize, unsigned osize, BOOL flip);

void
CreateConstantNode (unsigned value);

Node
getLastUnused (void);

Node
getLastInserted (void);

void
CreateStoreNode (Storage storage, unsigned isize);

void
CreateLoadNode (Storage storage, unsigned size);

void
TreeDump (Node node, FILE * file, char c, unsigned indent);

void
DumpNodes (FILE * file);

void
DumpNode (Node node, FILE * file);

void
OptimizeNodes (void);


#endif

/*===================================================================*\
|*	$Id: tree.h,v 1.1 2018/12/21 00:15:19 wolf Exp $
\*===================================================================*/
